# Copyright 2023 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Full test of builder_status module."""

from recipe_engine import post_process
from PB.recipe_modules.fuchsia.builder_status.tests.full import InputProperties

DEPS = [
    "fuchsia/buildbucket_util",
    "fuchsia/builder_status",
    "recipe_engine/buildbucket",
    "recipe_engine/properties",
    "recipe_engine/step",
]

PROPERTIES = InputProperties


def RunSteps(api, props):
    builder = api.builder_status.retrieve(
        bucket="ci",
        builder="test",
        n=10,
        include_incomplete=props.include_incomplete,
    )
    with api.step.nest("builder") as pres:
        pres.step_summary_text = repr(builder)

    if builder.exists:
        api.step("exists", [])
    else:
        api.step("does not exist", [])

    if builder.incomplete_included:
        if api.builder_status.is_incomplete(builder):
            api.step("is incomplete", [])
    else:
        api.step("skipping incomplete", [])

    if api.builder_status.is_passing(builder):
        api.step("is passing", [])

    if api.builder_status.has_recently_passed(builder):
        api.step("has recently passed", [])

    with api.step.nest("link") as pres:
        pres.links["link"] = builder.link


def GenTests(api):  # pylint: disable=invalid-name
    def props(include_incomplete=True):
        return api.properties(include_incomplete=include_incomplete)

    def assert_exists():
        return api.post_process(post_process.MustRun, "exists") + api.post_process(
            post_process.DoesNotRun, "does not exist"
        )

    def assert_does_not_exist():
        return api.post_process(post_process.DoesNotRun, "exists") + api.post_process(
            post_process.MustRun, "does not exist"
        )

    def assert_skipped_incomplete():
        return api.post_process(post_process.MustRun, "skipping incomplete")

    def assert_incomplete():
        return api.post_process(post_process.MustRun, "is incomplete")

    def assert_not_incomplete():
        return api.post_process(post_process.DoesNotRun, "is incomplete")

    def assert_passing():
        return api.post_process(post_process.MustRun, "is passing")

    def assert_not_passing():
        return api.post_process(post_process.DoesNotRun, "is passing")

    def assert_recently_passed():
        return api.post_process(post_process.MustRun, "has recently passed")

    def assert_not_recently_passed():
        return api.post_process(post_process.DoesNotRun, "has recently passed")

    def drop_expectations_must_be_last():
        # No need for expectation files, everything of note here is tested by
        # assertions. This must be the last thing added to the test.
        return api.post_process(post_process.DropExpectation)

    yield (
        api.buildbucket_util.test("no_buildbucket_config")
        + props()
        + assert_exists()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("bucket_does_not_exist")
        + props()
        + api.builder_status.buildbucket_config(foo=("bar", "baz"))
        + assert_does_not_exist()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("builder_does_not_exist")
        + props()
        + api.builder_status.buildbucket_config(ci=("bar", "baz"))
        + assert_does_not_exist()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("none")
        + props()
        + api.builder_status.buildbucket_config()
        + api.buildbucket.simulated_search_results([])
        + assert_exists()
        + assert_not_incomplete()
        + assert_not_passing()
        + assert_recently_passed()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("one_pass")
        + props(include_incomplete=False)
        + api.builder_status.buildbucket_config()
        + api.buildbucket.simulated_search_results([api.builder_status.passed()])
        + assert_exists()
        + assert_skipped_incomplete()
        + assert_passing()
        + assert_recently_passed()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("one_scheduled_one_incomplete_one_pass")
        + props()
        + api.builder_status.buildbucket_config()
        + api.buildbucket.simulated_search_results(
            [
                api.builder_status.scheduled(),
                api.builder_status.running(),
                api.builder_status.passed(),
            ]
        )
        + assert_exists()
        + assert_incomplete()
        + assert_passing()
        + assert_recently_passed()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("one_fail")
        + props()
        + api.builder_status.buildbucket_config()
        + api.buildbucket.simulated_search_results([api.builder_status.failure()])
        + assert_exists()
        + assert_not_incomplete()
        + assert_not_passing()
        + assert_not_recently_passed()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("one_scheduled_one_incomplete_one_fail")
        + props()
        + api.builder_status.buildbucket_config()
        + api.buildbucket.simulated_search_results(
            [
                api.builder_status.scheduled(),
                api.builder_status.running(),
                api.builder_status.failure(),
            ]
        )
        + assert_exists()
        + assert_incomplete()
        + assert_not_passing()
        + assert_not_recently_passed()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("one_pass_one_fail")
        + props()
        + api.builder_status.buildbucket_config()
        + api.buildbucket.simulated_search_results(
            [api.builder_status.passed(), api.builder_status.failure()]
        )
        + assert_exists()
        + assert_not_incomplete()
        + assert_passing()
        + assert_recently_passed()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("one_fail_one_pass")
        + props()
        + api.builder_status.buildbucket_config()
        + api.buildbucket.simulated_search_results(
            [api.builder_status.failure(), api.builder_status.passed()]
        )
        + assert_exists()
        + assert_not_incomplete()
        + assert_not_passing()
        + assert_recently_passed()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("many_passes")
        + props()
        + api.builder_status.buildbucket_config()
        + api.buildbucket.simulated_search_results([api.builder_status.passed()] * 10)
        + assert_exists()
        + assert_not_incomplete()
        + assert_passing()
        + assert_recently_passed()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("many_failures")
        + props()
        + api.builder_status.buildbucket_config()
        + api.buildbucket.simulated_search_results([api.builder_status.failure()] * 10)
        + assert_exists()
        + assert_not_incomplete()
        + assert_not_passing()
        + assert_not_recently_passed()
        + drop_expectations_must_be_last()
    )

    yield (
        api.buildbucket_util.test("one_infra")
        + props()
        + api.builder_status.buildbucket_config()
        + api.buildbucket.simulated_search_results([api.builder_status.infra_failure()])
        + assert_exists()
        + assert_not_incomplete()
        + assert_not_passing()
        + assert_not_recently_passed()
        + drop_expectations_must_be_last()
    )
