# Copyright 2021 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from typing import Generator

from google.protobuf import json_format as jsonpb

from PB.recipe_modules.fuchsia.cipd_util.upload_manifest import CIPDUploadManifest
from PB.recipes.fuchsia.bazel_build_test_upload import BuilderManifest, InputProperties
from recipe_engine import recipe_api, recipe_test_api

DEPS = [
    "fuchsia/buildbucket_util",
    "fuchsia/cipd_util",
    "fuchsia/git_checkout",
    "recipe_engine/context",
    "recipe_engine/file",
    "recipe_engine/path",
    "recipe_engine/properties",
    "recipe_engine/proto",
    "recipe_engine/step",
]

PROPERTIES = InputProperties


def RunSteps(api: recipe_api.RecipeScriptApi, props: InputProperties):
    # Projects should not assume they'll always be checked out at the same
    # absolute path, so do the checkout in a random temporary directory.
    checkout_dir = api.path.mkdtemp("checkout")
    _, git_revision = api.git_checkout(props.remote, path=checkout_dir)

    build_dir = api.path.mkdtemp("build")
    with api.context(cwd=checkout_dir):
        builder_manifest = api.file.read_proto(
            "read builder manifest",
            "manifest.json",
            BuilderManifest,
            codec="JSONPB",
        )

        cipd_manifests = [
            api.file.read_proto(
                f"read cipd manifest {cipd_manifest_path}",
                checkout_dir / cipd_manifest_path,
                CIPDUploadManifest,
                codec="JSONPB",
            )
            for cipd_manifest_path in builder_manifest.cipd_manifests
        ]

        cas_digests = {}
        for cipd_manifest in cipd_manifests:
            step_name = f"cipd upload {cipd_manifest.pkg_name}"
            if not props.upload_to_cipd:
                step_name += " (dry run)"
            with api.step.nest(step_name):
                api.cipd_util.upload_from_manifest(
                    cipd_manifest.pkg_name,
                    cipd_manifest,
                    build_dir=build_dir,
                    repository=props.remote,
                    git_revision=git_revision,
                    upload_to_cipd=props.upload_to_cipd,
                    cas_digests=cas_digests,
                )


def GenTests(api) -> Generator[recipe_test_api.TestData, None, None]:
    default_remote = "https://fuchsia.googlesource.com/foo"

    def properties(**kwargs):
        props = {
            "remote": default_remote,
            "build_target": "//src:infra",
            "build_options": ["--config", "fuchsia_x64"],
            "default_service_account": "artifact-readers@fuchsia-infra.iam.gserviceaccount.com",
        }
        props.update(kwargs)
        return api.properties(jsonpb.ParseDict(props, InputProperties()))

    def cipd_manifest_data(files=None):
        if not files:
            files = {"local/foo/bar": "package/foobar"}
        return api.proto.output(
            CIPDUploadManifest(
                pkg_name="fuchsia/foo",
                files=[
                    CIPDUploadManifest.FileToUpload(source=k, dest=v)
                    for k, v in files.items()
                ],
            )
        )

    yield (
        api.buildbucket_util.test(
            "invalid_cipd_upload",
            git_repo=default_remote,
            # The manifest should be validated even in presubmit.
            tryjob=True,
            status="FAILURE",
        )
        + properties()
        + api.step_data(
            "read builder manifest",
            api.file.read_proto(
                BuilderManifest(
                    cipd_manifests=["cipd_manifest.json"],
                ),
            ),
        )
        + api.step_data(
            "read cipd manifest cipd_manifest.json",
            cipd_manifest_data(
                # Paths outside the build directory are disallowed.
                {"foo/../../checkout-path": "package_path/foo/bar"}
            ),
        )
    )

    yield (
        api.buildbucket_util.test(
            "cipd_upload_tryjob", git_repo=default_remote, tryjob=True
        )
        + properties(
            external_presubmit_options={
                "gerrit_host": "foo-review.googlesource.com",
                "gerrit_project": "bar",
            },
        )
        + api.step_data(
            "read builder manifest",
            api.file.read_proto(
                BuilderManifest(
                    cipd_manifests=["cipd_manifest.json"],
                ),
            ),
        )
        + api.step_data(
            "read cipd manifest cipd_manifest.json",
            cipd_manifest_data(),
        )
    )

    yield (
        api.buildbucket_util.test("mos_cipd", git_repo=default_remote)
        + properties(upload_to_cipd=True)
        + api.step_data(
            "read builder manifest",
            api.file.read_proto(
                BuilderManifest(
                    cipd_manifests=["cipd_manifest.json"],
                ),
            ),
        )
        + api.step_data(
            "read cipd manifest cipd_manifest.json",
            cipd_manifest_data(
                # Cover the case where we upload an entire subdirectory into the
                # root of the CIPD package.
                {"dir/pkg": "."},
            ),
        )
    )
