# Copyright 2013 The Chromium Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from recipe_engine.recipe_api import Property

DEPS = [
    "fuchsia/git",
    "recipe_engine/buildbucket",
    "recipe_engine/context",
    "recipe_engine/file",
    "recipe_engine/path",
    "recipe_engine/properties",
    "recipe_engine/step",
    "recipe_engine/time",
]

PROPERTIES = {
    "path": Property(
        kind=str, help="Directory to clone into, relative to start_dir", default=None
    ),
    "branch": Property(kind=str, help="Git branch", default=None),
    "revision": Property(kind=str, help="Revision", default=None),
    "snap_ref": Property(kind=str, help="Commit hash ref to snap to", default=None),
    "snap_path": Property(
        kind=str,
        help="Directory to checkout snap in, relative to start_dir",
        default=None,
    ),
}

URL = "https://fuchsia.googlesource.com/fuchsia"


def RunSteps(
    api,
    path,
    branch,
    revision,
    snap_ref,
    snap_path,
):
    absolute_path = api.path.start_dir / path if path else None

    root_dir = absolute_path or api.path.start_dir / "fuchsia"

    api.git.remote_alias(URL + ".git")

    with api.context(cwd=root_dir):
        api.git.init(bare=True)
        api.git.fetch(
            URL,
            refspec=revision,
            recurse_submodules=True,
            prune=True,
            tags=True,
            depth=5,
            filters=True,
        )
        api.git.fetch(
            all_remotes=True,
            recurse_submodules=False,
            filters="blob:limit=1k",
        )
        api.git.get_hash()
        api.git.remote_add("origin", URL)
        api.git.get_remote_url("origin")
        api.git.get_timestamp()
        api.git.get_commit_message()
        api.git.get_commit_message(oneline=True)
        api.git.get_changed_files(
            test_data=["foo.bar", "bar.foo"],
            base="origin/main",
        )
        api.git.config_replace_all("foo", "bar", r".*")
        api.git.clean(force=True, recursive=True, ignore_rules=True)
        api.git.sync_submodule(recursive=True)

        # You can invoke arbitrary command on api.git.
        api.git("git status", "status", config={"foo": "bar"})

        # List branches.
        api.git.branch(contains="foo")
        api.git.branch(step_name="git branch remotes", contains="foo", remotes=True)

        # You can use api.git.rebase to rebase the current branch onto another one
        api.git.rebase("origin/main", rebase_merges=True)

        # You can use api.git.describe to describe a commit by its tag(s).
        api.git.describe(commit="foo", tags=True, contains=True)

        # You can also specify your expected number of tags.
        api.git.describe(commit="foo", tags=True, contains=True, expected_num=2)
        try:
            api.git.describe(commit="foo", tags=True, contains=True, expected_num=1)
        except api.step.StepFailure:
            pass

        # Get default remote branch and branch head.
        remote_branch = api.git.get_default_remote_branch(url=URL)
        remote_branch = remote_branch.replace("refs/heads/", "")
        api.git.get_remote_branch_head(url=URL, branch=remote_branch)

        # Get all remote branch heads.
        api.git.get_all_remote_branch_heads(url=URL)

        # Get remote tag commit if it exists.
        api.git.get_remote_tag(url=URL, tag="testtag")

        # Add a new file.
        api.file.write_text("drop file", root_dir / "time.txt", str(api.time.time()))
        api.git.add(pathspec=root_dir / "time.txt", only_tracked=True)

        # Add with options.
        api.git.add(add_all=True, intent_to_add=True, force=True)

        # Remove files from index recursively.
        api.git.rm(pathspec=root_dir, recursive=True)

        # Restore files.
        api.git.restore(pathspec=root_dir)

        # Show files.
        api.git.ls_files(
            modified=True,
            deleted=True,
            others=True,
            exclude_standard=True,
            line_termination=True,
            file="*.txt",
        )

        # Update specified submodules from remote.
        api.git.update_submodule(
            paths=["a", "b", "c"], recursive=False, remote=True, filters=True
        )
        api.git.update_submodule(
            paths=["a", "b", "c"],
            filters=("blob:limit=1k", "object:type=blob"),
            fetch=False,
        )

        # Update all submodules recursively.
        api.git.update_submodule(recursive=True, force=True)

        # Check submodule status.
        api.git.submodule_status()

        # Commit the change
        api.git.commit("example change 1", no_verify=True)

        # Commit the change with the file named explicitly
        api.git.commit("example change 2", files=[root_dir / "time.txt"])

        # Commit the change with all tracked files
        api.git.commit("example change 3", all_tracked=True)

        # Commit the change with all tracked and untracked files
        api.git.commit("example change 4", all_files=True)

        # Amend the change
        api.git.commit("example change 4 (amended)", amend=True)

        # Amend the change with no edit
        api.git.commit(no_edit=True, amend=True)

        # Commit the change, overriding the committer
        api.git.commit(
            "example change 5",
            author_override={"name": "foo", "email": "bar@example.com"},
            allow_empty=True,
        )

        # Get author name of the commit
        api.git.get_author_name(commit="abcde")

        # Get author email of the commit
        api.git.get_author_email(commit="abcde")

        # check if commit_1 is ancestor of commit_2
        api.git.is_ancestor(commit_1="abcdef", commit_2="ghjklm")

        # Check if commit_1 is ancestor of the current HEAD.
        api.git.is_ancestor(commit_1="abcdef")

        # Tag the commit.
        api.git.tag("v1.0.0")

        # Cherry-pick a commit.
        api.git.cherry_pick("abcdef", strategy_option="theirs")

        # Revert the commit.
        api.git.revert("abcdef", strategy_option="ours")

        patch_file = api.path.tmp_base_dir / "git.patch"
        # Take the diff.
        api.git.diff()

        # Take the diff between the last 2 commits with extra options.
        api.git.diff(
            ref_base="HEAD^",
            ref="HEAD",
            patch_file=patch_file,
            unified=5,
            cached=True,
            exit_code=True,
        )

        # Do a raw checkout.
        api.git.raw_checkout(ref="HEAD", force=True)

        # Compute object ID.
        api.git.hash_object("foo")

        # Get status of the working directory and the staging area.
        api.git.status()

        # Get log.
        api.git.log(fmt="%H")

        # Apply a patch.
        api.git.apply_patchfile(patch=patch_file)

        api.git.upload_cl(
            labels={
                "Label": 1,
            },
            cl_notify_option="NONE",
            cc_emails=["foo@google.com"],
            reviewer_emails=["bar@google.com"],
        )

        # Atomically push multiple refs with additional options.
        api.git.push(
            ["HEAD:refs/heads/main", "refs/tags/foo"],
            atomic=True,
            dryrun=True,
            options=["nokeycheck"],
        )

        api.git.config_remove_section(section="feature.experimental")

        # Count the number of revisions between two arbitrary revisions.
        api.git.rev_list_count("foo", "bar", test_data="4")

        # Count the number of revisions to an arbitrary revision.
        api.git.rev_list_count("foo", test_data="1")

        with api.step.nest("remotes") as pres:
            for rem in api.git.get_remotes():
                pres.links[rem.name] = rem.url

    assert api.git.read_commit_footer("Foo\n\nBar: Iabc", "Bar") == "Iabc"
    assert api.git.read_commit_footer("Foo\n\nBar: 123", "No-Such-Footer") is None

    if snap_ref:
        # Snap branch to snap ref.
        api.git.snap_branch(
            snap_ref=snap_ref,
            branch=branch,
            message="example snap",
            path=api.path.start_dir / snap_path if snap_path else None,
            push=True,
        )

    api.git.list_submodules(
        "\n".join(
            [
                "submodule.third_party/foobar.path=third_party/foobar",
                "submodule.third_party/foobar.url=https://fuchsia.googlesource.com/third_party/github.com/foobar/foobar",
                "submodule.third_party/foobar.unknown_field=foobar",
                "submodule.build/secondary/third_party/foobaz.path=build/secondary/third_party/foobaz",
                "submodule.build/secondary/third_party/foobaz.url=https://fuchsia.googlesource.com/foobaz-gn",
                "submodule.build/secondary/third_party/foobaz.unknown_field=foobaz",
            ]
        )
    )


def GenTests(api):
    yield api.test("basic") + api.git.get_changed_files(
        "git diff-tree", ["foo.txt"]
    ) + api.git.rev_parse("git rev-parse", "foo")
    yield api.test("basic_ci") + api.buildbucket.ci_build(
        git_repo=URL
    ) + api.git.branch("git branch", ["a", "b"]) + api.git.branch(
        "git branch remotes", ["c", "d"], remotes=True
    )
    yield api.test("basic_cq") + api.buildbucket.try_build(git_repo=URL)
    yield api.test("basic_path") + api.properties(path="foo")
    yield api.test("basic_ref") + api.properties(revision="refs/foo/bar")
    yield api.test("basic_branch") + api.properties(
        revision="refs/heads/testing"
    ) + api.git.get_remote_branch_head(
        "git ls-remote", "foobar"
    ) + api.git.get_all_remote_branch_heads(
        "git ls-remote",
        {
            "refs/heads/main": "123",
            "refs/heads/extra_branch": "456",
        },
    )
    yield api.test("basic_hash") + api.properties(
        revision="abcdef0123456789abcdef0123456789abcdef01"
    )
    yield api.test("snap") + api.properties(
        snap_ref="5n4p",
        branch="snap_branch",
    )
    yield api.test("snap_path") + api.properties(
        snap_ref="5n4p",
        branch="snap_branch",
        snap_path="snap_workdir",
    )
    yield (
        api.test("snap_to_same_ref")
        + api.properties(
            snap_ref="deadbeef",
            branch="snap_branch",
            **{"$fuchsia/git": {"trace": True}},
        )
    )
