# Copyright 2015 The Chromium Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from recipe_engine.recipe_api import Property

DEPS = [
    "fuchsia/jiri",
    "recipe_engine/context",
    "recipe_engine/path",
    "recipe_engine/platform",
    "recipe_engine/properties",
    "recipe_engine/raw_io",
]

PROPERTIES = {
    "override": Property(
        kind=bool,
        help="Whether to override the imported manifest with "
        + "commit's given revision.",
        default=False,
    ),
}


def RunSteps(api, override):
    # Setup jiri root with no submodules.
    api.jiri.init(
        "dir",
        enable_submodules=False,
        use_lock_file=True,
        attributes=["a", "b"],
        dissociate=True,
    )

    # Setup a new jiri root with submodules enabled.
    api.jiri.init(
        "dir",
        enable_submodules=True,
        use_lock_file=True,
        attributes=["a", "b"],
        dissociate=True,
    )

    # Get jiri executable.
    assert api.jiri.jiri

    # Import the manifest.
    api.jiri.import_manifest(
        "minimal",
        "https://fuchsia.googlesource.com",
        name="manifest",
        revision="c22471f4e3f842ae18dd9adec82ed9eb78ed1127",
        overwrite=True,
    )

    # Import the manifest with a remote branch.
    api.jiri.import_manifest(
        "minimal",
        "https://fuchsia.googlesource.com",
        name="manifest",
        revision="c22471f4e3f842ae18dd9adec82ed9eb78ed1127",
        remote_branch="main",
        overwrite=True,
    )

    if override:
        api.jiri.override(
            project="fuchsia",
            remote="https://fuchsia.googlesource.com/fuchsia",
            new_revision="abcd1234",
        )

    # Download all projects.
    with api.context(cwd=api.path.start_dir):
        api.jiri.update(
            gc=True,
            snapshot="snapshot",
            local_manifest=True,
            override_optional=True,
            packages_to_skip=("a", "b"),
            timeout=10 * 60,
        )

    # Get the revision of the imported manifest.
    api.jiri.get_import_revision("manifest", name="get import revision", ok_ret="any")

    # Edit manifest.
    api.jiri.edit_manifest(
        "minimal",
        projects=["a", ("b", "c22471f4e3f842ae18dd9adec82ed9eb78ed1127")],
        packages=[("e", "z1R_-lysycEAb35RDsLgM0YfqCwIT4zvu5_lCVPV37gC")],
    )

    # Resolve CIPD lock file.
    api.jiri.resolve(
        local_manifest=True,
        enable_project_lock=True,
        enable_package_lock=True,
        output=api.path.cleanup_dir,
        manifests=["minimal"],
    )

    # Read a project from the manifest.
    api.jiri.read_manifest_element(
        manifest="minimal",
        element_type="project",
        element_name="test/project",
    )

    # Read a package from the manifest.
    api.jiri.read_manifest_element(
        manifest="minimal",
        element_type="package",
        element_name="test/package",
    )

    # Run update with submodule rebase.
    api.jiri.update(rebase_submodules=True, run_hooks=False)

    # Run hooks separately.
    api.jiri.update(rebase_tracked=True, run_hooks=False)
    with api.context(cwd=api.path.start_dir):
        api.jiri.run_hooks(
            local_manifest=True, fetch_packages=False, hook_timeout_secs=60
        )
        api.jiri.run_hooks(packages_to_skip=("a", "b"))

    # Fetch packages separately.
    with api.context(cwd=api.path.start_dir):
        api.jiri.update(rebase_tracked=True, fetch_packages=False)
    api.jiri.fetch_packages(local_manifest=True, packages_to_skip=("a", "b"))

    # Tals_ke a snapshot.
    api.jiri.snapshot(api.path.start_dir / "test/snapshot")

    # Take additional cipd ensure and version snapshots.
    api.jiri.snapshot(api.path.start_dir / "test/snapshot", cipd=True)

    # Take diff between snapshots.
    snapshot_file_1 = api.path.mkstemp()
    snapshot_file_2 = api.path.mkstemp()
    api.jiri.snapshot(snapshot_file=snapshot_file_1)
    api.jiri.snapshot(snapshot_file=snapshot_file_2)
    api.jiri.snapshot_diff(snapshot_1=snapshot_file_1, snapshot_2=snapshot_file_2)

    # Generate source manifest.
    api.jiri.source_manifest()

    # Get information about the project.
    api.jiri.project(["test"])

    # Get information about the project based on local manifest
    api.jiri.project(["read-local-manifest"], local_manifest=True)

    # Get information about a remote project.
    api.jiri.project(["remote-test-project"], list_remote=True)

    # Get information about a package.
    api.jiri.package(["test/package"])

    # Patch in an existing change.
    api.jiri.patch(
        "refs/changes/1/2/3",
        host="https://fuchsia-review.googlesource.com",
        project="test",
        delete=True,
        force=True,
    )

    # Reset to JIRI_HEAD.
    api.jiri.reset()

    api.jiri.generate_gitmodules(
        ".gitmodules", generate_script="foo.py", redirect_root=True
    )

    api.jiri.check_clean()
    api.jiri.clean()


def GenTests(api):
    yield (
        api.test("basic_ci")
        + api.jiri.read_manifest_element(
            "test/project",
            test_output={
                "gerrithost": "project_gerrit_host",
                "githooks": "project_githooks",
                "historydepth": "project_historydepth",
                "name": "project_name",
                "path": "project_path",
                "remote": "project_remote",
                "revision": "project_revision",
            },
        )
        + api.step_data("get import revision", retcode=1)
    )

    yield api.test("basic_ci_mac") + api.platform.name(name="mac")

    yield api.test("ci_with_override") + api.properties(override=True)

    # This should result in an infra failure.
    yield (
        api.test("patch_error", status="FAILURE")
        + api.step_data(
            "jiri patch",
            api.raw_io.stream_output_text(
                "ERROR: patch failed", stream="stderr", retcode=1
            ),
        )
    )

    # Either of these patch failures should result in a non-infra failure.
    for name, exc_type in {
        "patch_no_such_project": api.jiri.NoSuchProjectError,
        "patch_rebase_error": api.jiri.RebaseError,
    }.items():
        yield (
            api.test(name, status="FAILURE")
            + api.step_data("jiri patch", retcode=exc_type.EXIT_CODE)
        )

    # This should result in a non-infra failure.
    yield (
        api.test("cipd_ensure_file_verify_error_runhooks", status="FAILURE")
        + api.step_data(
            "jiri run-hooks", retcode=api.jiri.EnsureFileVerificationError.EXIT_CODE
        )
    )

    # This should result in a non-infra failure.
    yield (
        api.test("cipd_ensure_file_verify_error_fetch_packages", status="FAILURE")
        + api.step_data(
            "jiri fetch-packages",
            retcode=api.jiri.EnsureFileVerificationError.EXIT_CODE,
        )
    )

    yield (
        api.test("example_resolve_with_platform")
        + api.step_data(
            "jiri resolve",
            stdout=api.jiri.example_resolve_data("public/package/${platform}"),
        )
    )
    yield (
        api.test("example_resolve_no_platform")
        + api.step_data(
            "jiri resolve", stdout=api.jiri.example_resolve_data("public/package")
        )
    )
    yield (
        api.test("missing_packages", status="INFRA_FAILURE")
        + api.step_data("jiri package", api.jiri.package([]))
    )
