# Copyright 2020 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from recipe_engine.config import List
from recipe_engine.recipe_api import Property

DEPS = [
    "fuchsia/presubmit_util",
    "recipe_engine/buildbucket",
    "recipe_engine/properties",
]


PROPERTIES = {
    "trigger_tryjobs": Property(
        kind=bool,
        default=False,
        help="Whether to trigger tryjobs explicitly.",
    ),
    "tryjobs": Property(
        kind=List(str),
        default=None,
        help="Tryjobs to trigger.",
    ),
    "use_gclient_override": Property(
        kind=bool,
        default=False,
        help="Whether to use gclient override.",
    ),
    "use_package_overrides": Property(
        kind=bool,
        default=False,
        help="Whether to use package overrides.",
    ),
}


def RunSteps(
    api,
    trigger_tryjobs,
    tryjobs,
    use_gclient_override,
    use_package_overrides,
):
    options = api.presubmit_util.Options(
        gerrit_host="fuchsia-review.googlesource.com",
        gerrit_project="infra/test",
        ref="refs/heads/main",
        timeout_secs=1800,
        tryjobs=tryjobs,
        tryjobs_gerrit_host="alternate-review.googlesource.com",
        trigger_tryjobs=trigger_tryjobs,
        tryjobs_wait_secs=60,
        presubmit_status_gerrit_label="Verified",
    )
    api.presubmit_util.orchestrate(
        options=options,
        cl_subject="[infra] Very good code",
        file_edits=[("path/to/a.txt", "foo"), ("path/to/b.txt", "bar")],
        package_overrides={"package": "abcdef"} if use_package_overrides else None,
        gclient_variables={"override": "foobar"} if use_gclient_override else None,
    )


def GenTests(api):
    create_cl_test_data = api.presubmit_util.create_cl()
    tryjobs = ["fuchsia/try/foo.x64-debug", "fuchsia/try/foo.arm64-debug"]
    tryjob_msgs = [
        api.buildbucket.try_build_message(
            project="fuchsia",
            bucket="try",
            builder="bar.x64-debug",
        ),
    ]
    yield (
        api.test("gclient_override_with_explicit_tryjobs")
        + api.properties(
            use_gclient_override=True, tryjobs=tryjobs, trigger_tryjobs=True
        )
        + api.presubmit_util.trigger_tryjobs(
            tryjob_msgs=tryjob_msgs,
        )
        + api.presubmit_util.collect_tryjobs(
            tryjob_msgs=tryjob_msgs,
        )
    )
    yield (
        api.test("package_overrides")
        + create_cl_test_data
        + api.presubmit_util.wait_for_cq()
        + api.properties(use_package_overrides=True)
    )
    yield (
        api.test("failed_cq", status="FAILURE")
        + create_cl_test_data
        + api.presubmit_util.wait_for_cq(success=False)
    )
    yield (
        api.test("failed_tryjobs", status="FAILURE")
        + create_cl_test_data
        + api.properties(tryjobs=tryjobs)
        + api.presubmit_util.collect_tryjobs(builders=tryjobs, status="FAILURE")
    )
    yield (
        api.test("missing_tryjob", status="INFRA_FAILURE")
        + create_cl_test_data
        + api.properties(tryjobs=tryjobs)
    )
