# Copyright 2019 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from recipe_engine.recipe_api import Property

DEPS = [
    "fuchsia/release",
    "recipe_engine/path",
    "recipe_engine/properties",
    "recipe_engine/raw_io",
    "recipe_engine/step",
]

PROPERTIES = {
    "ref": Property(kind=str, help="Git ref to search release versions on."),
    "date": Property(kind=str, help="Release date as YYYYMMDD."),
}


def RunSteps(api, ref, date):
    repo_path = api.path.start_dir
    if not api.release.ref_to_release_version(ref=ref, repo_path=repo_path):
        raise api.step.StepFailure("not a release version")
    api.release.get_branches(repo_path=repo_path)
    with api.step.nest("parent step") as presentation:
        m = api.release.get_major_number(repo_path=repo_path)
        api.release.write_major_number(m + 1, repo_path=repo_path)
        api.release.ref_to_release_version(ref=ref, repo_path=repo_path)
        release_branches = api.release.get_branches(repo_path=repo_path)
        presentation.properties["release_branches"] = release_branches
        api.release.set_output_properties(
            presentation,
            ref,
            api.release.ReleaseVersion.from_string("0.20191019.1.2"),
            "https://fuchsia.googlesource.com/integration",
        )
    api.release.get_initial_release_version(date=date, repo_path=repo_path)
    api.release.get_next_release_version(ref=ref, date=date, repo_path=repo_path)
    api.release.get_next_candidate_version(ref=ref, repo_path=repo_path)
    api.release.get_next_candidate_version(ref=ref, add_track=True, repo_path=repo_path)
    api.release.get_release_versions(ref=ref, date=date, repo_path=repo_path)
    api.release.get_next_candidate_version(ref="not_a_tag", repo_path=repo_path)
    api.release.get_current_release_version(repo_path=repo_path)
    api.release.lookup_builds(
        builders=["a/b/c", "d/e/f"],
        revision=ref,
        remote="https://fuchsia.googlesource.com/integration",
    )
    api.release.shorten_revision(revision=ref)
    assert api.release.validate_branch("releases/abc")
    assert not api.release.validate_branch("releases/abc/def")
    assert api.release.get_milestone_from_branch("releases/f12", "f") == 12


def GenTests(api):
    test_tags = [
        "releases/0.20191019.1.2",
        "releases/0.20191019.1.1",
        "releases/0.20191019.0.1",
        "releases/0.20191018.0.1",
        "releases/0.20191017.1.10",
        "not_a_release_version",
        "releases/0.20191017.1.9",
    ]
    yield (
        api.test("basic")
        + api.properties(date="20191019", ref="foo")
        + api.release.ref_to_release_version("releases/0.20191019.1.1")
        + api.release.get_branches(["main", "releases/foo", "releases/bar"])
        + api.release.ref_to_release_version(
            "releases/0.20191019.1.1", nesting="parent step"
        )
        + api.release.get_branches(
            ["main", "releases/foo", "releases/bar"], nesting="parent step"
        )
        + api.release.ref_to_release_version("releases/0.20191019.1.1", rep=2)
        + api.release.ref_to_release_version("releases/0.20191019.1.1", rep=3)
        + api.release.ref_to_release_version("releases/0.20191019.1.1", rep=4)
        + api.release.ref_to_release_version("releases/0.20191019.1.1", rep=5)
        + api.release.ref_to_release_version("", rep=4, retcode=1)
        + api.step_data(
            "get release versions on foo",
            api.raw_io.stream_output_text("\n".join(test_tags)),
        )
        + api.step_data(
            "get release versions on foo (2)",
            api.raw_io.stream_output_text("\n".join(test_tags)),
        )
        + api.release.get_major_number(1)
        + api.release.get_major_number(2, nesting="parent step")
    )
    yield (
        api.test("not_a_release_version", status="FAILURE")
        + api.properties(date="20191019", ref="foo")
        + api.release.ref_to_release_version("not_a_release_version")
    )
