# Copyright 2023 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import collections

from recipe_engine import recipe_api

from PB.go.fuchsia.dev.shac_project.shac.internal.sarif import sarif as sarif_pb2

from RECIPE_MODULES.fuchsia.utils import pluralize

# The Gerrit API puts a limit on the maximum total comments and size of comments
# that a single change can have, so it's important to not emit too many
# comments. Too many comments will probably just be annoying anyway.
MAX_COMMENTS = 50


class ShacApi(recipe_api.RecipeApi):
    def check(
        self,
        root,
        shac_path,
        revision,
        entrypoint=None,
        runtime_vars=None,
        show_repro_cmd=True,
        emit_comments=True,
        test_id_prefix=None,
    ):
        """Runs `shac check` and uploads results to ResultDB and Gerrit.

        Args:
            root (Path): Path to the root of the repository to analyze.
            shac_path (Path): Path to a shac executable to use.
            revision (str): git revision of the repository being analyzed.
            entrypoint (str): Basename of shac files to run, if not "shac.star".
            runtime_vars (dict): Runtime variables to pass into shac.
            show_repro_cmd (bool): Whether to show `shac` local repro command in
              the step UI.
            emit_comments (bool): Whether to emit comments to Gerrit.
            test_id_prefix (str): Optional prefix to add to ResultDB test IDs.
        """
        cmd = [
            shac_path,
            "check",
            "--json-output",
            self.m.proto.output(sarif_pb2.Document, "JSONPB"),
        ]
        repro_cmd = [
            "shac",
            "check",
        ]
        common_args = [
            # Run checks on all files, not just affected files. New static
            # analysis errors can appear in unaffected files on changes that
            # affect shac Starlark files or any scripts/tools that shac
            # interfaces with, so it can't be assumed that new findings can only
            # show up in affected files.
            "--all",
        ]
        if entrypoint:
            common_args.extend(["--entrypoint", entrypoint])
        for k, v in sorted((runtime_vars or {}).items()):
            assert isinstance(k, str), f"runtime var keys must be strings: {k!r}"
            assert isinstance(v, str), f"runtime var values must be strings: {v!r}"
            cmd.extend(["--var", f"{k}={v}"])

        cmd.extend(common_args)
        repro_cmd.extend(common_args)

        # Set the working directory via api.context instead of via shac's -C
        # flag in case the executable specified by `shac_path` is actually a
        # wrapper script with logic that assumes it's run from within the
        # checkout root.
        with self.m.context(cwd=root):
            step = self.m.step(
                "shac check",
                self.m.resultdb.wrap(cmd, test_id_prefix=test_id_prefix),
                ok_ret=(0, 1),
                step_test_data=lambda: self.m.proto.test_api.output(
                    sarif_pb2.Document()
                ),
            )
        if show_repro_cmd:
            step.presentation.step_text = (
                f"Run `{' '.join(repro_cmd)}` to reproduce locally."
            )
        if step.retcode:
            step.presentation.status = self.m.step.FAILURE

        comments = []
        # Mapping from check name to list of error messages.
        errors_by_check = collections.defaultdict(list)

        for run in step.proto.output.runs:
            check_name = run.tool.driver.name
            for result in run.results:
                # Schema:
                # https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#robot-comment-input
                comment = {
                    "robot_id": check_name,
                    "robot_run_id": f"{revision}/{check_name}",
                    "message": result.message.text,
                }
                if result.locations:
                    loc = result.locations[0].physical_location
                    path = loc.artifact_location.uri or "/COMMIT_MSG"
                    # If start_line equals 0, then the whole region is zeroes, which
                    # indicates the entire file. Gerrit applies the comment to the
                    # entire file if `range` is unset.
                    if loc.region.start_line > 0:
                        comment["range"] = self._sarif_region_to_gerrit_range(
                            path, loc.region, root
                        )
                else:
                    path = "/PATCHSET_LEVEL"
                if result.level == "error":
                    line = ""
                    if rng := comment.get("range"):
                        line = f"({rng['start_line']})"
                    # Only include the first line of the message, other lines
                    # will mess up formatting of the build summary.
                    msg = result.message.text.splitlines()[0]
                    errors_by_check[check_name].append(f"{path}{line}: {msg}")

                # Not used by the Gerrit API, but useful for pre-processing.
                comment["level"] = result.level

                comment["path"] = path
                comment["fix_suggestions"] = [
                    {
                        "description": fix.description.text,
                        "replacements": [
                            {
                                "path": change.artifact_location.uri,
                                "range": self._sarif_region_to_gerrit_range(
                                    change.artifact_location.uri,
                                    repl.deleted_region,
                                    root,
                                ),
                                "replacement": repl.inserted_content.text,
                            }
                            for change in fix.artifact_changes
                            for repl in change.replacements
                        ],
                    }
                    for fix in result.fixes
                ]
                comments.append(comment)

        with self.m.step.nest("get changed files"):
            changed_files = self._changed_files(root)

        # Only include comments for changed files or special files like
        # "/COMMIT_MESSAGE".
        comments = [
            c
            for c in comments
            if c["path"] in changed_files or c["path"].startswith("/")
        ]
        if len(comments) > MAX_COMMENTS:
            num_skipped = len(comments) - MAX_COMMENTS
            comments.sort(
                key=lambda comment: (
                    # Sort special comments first, followed by errors, since
                    # those are the most important comments to emit. Other
                    # comments are more acceptable to skip.
                    not comment["path"].startswith("/"),
                    comment["level"] != "error",
                    # Finally, sort by path so the sorting is completely
                    # deterministic.
                    comment["path"],
                ),
            )
            comments = comments[:MAX_COMMENTS]
            comments.append(
                {
                    "path": "/PATCHSET_LEVEL",
                    "robot_id": check_name,
                    "robot_run_id": f"{revision}/{check_name}",
                    "message": (
                        f"WARNING: too many comments to add, {pluralize('comment', num_skipped)} not emitted."
                    ).strip(),
                    "level": None,
                }
            )

        if emit_comments and comments:
            gerrit_change = self.m.buildbucket.build.input.gerrit_changes[0]

            # List the comments that have already been added by previous
            # shac runs to avoid emitting duplicate comments. This isn't
            # perfect: if two runs of the same builder run in sync, they may
            # race and both end up emitting comments. But that should be
            # rare, and it's not the end of the world if it does happen.
            existing = self.m.gerrit.list_robot_comments(
                "get previous robot comments",
                change_id=gerrit_change.change,
                revision_id=gerrit_change.patchset,
                max_attempts=3,
                test_data=self.m.json.test_api.output({}),
            ).json.output

            # Only some fields are relevant for determining whether a new
            # comment is identical to an existing comment.
            filter_fields = ("message", "robot_id", "range")

            def is_dupe(comment):
                for other in existing.get(comment["path"], []):
                    if all(
                        # Coerce falsy values to None because they're
                        # functionally the same.
                        (comment.get(f) or None) == (other.get(f) or None)
                        for f in filter_fields
                    ):
                        return True
                return False

            # Filter out comments that have already been added on the
            # current patchset.
            new_comments = collections.defaultdict(list)
            for c in comments:
                # level field is no longer needed, remove it before passing into the
                # Gerrit API.
                c.pop("level")
                if not is_dupe(c):
                    new_comments[c["path"]].append(c)

            if new_comments:
                self.m.gerrit.set_review(
                    "send robot comments",
                    change_id=gerrit_change.change,
                    revision=gerrit_change.patchset,
                    host=gerrit_change.host,
                    robot_comments=new_comments,
                )

        if not step.retcode:
            return

        msg = "Checks failed"
        if errors_by_check:
            msg += ":\n\n"
        for check, errors in sorted(errors_by_check.items()):
            msg += f"- {check}\n"
            max_errors_to_show = 10
            to_hide = len(errors) - max_errors_to_show
            if to_hide > 1:
                errors[max_errors_to_show:] = [f"({to_hide} more errors)"]
            for err in errors:
                msg += f"  - {err}\n"

        raise self.m.step.StepFailure(msg)

    def _sarif_region_to_gerrit_range(self, path, region, root):
        # If start_line is zero than all other fields are assumed to be zero, which
        # indicates the entire file.
        if region.start_line == 0:
            # Gerrit has no shortcut for regions that apply to the entire file, we
            # must read the whole file to determine what the end line and column
            # should be.
            lines = self.m.file.read_text(f"read {path}", root / path).split("\n")
            region = sarif_pb2.Region(
                start_line=1,
                start_column=1,
                end_line=len(lines),
                end_column=len(lines[-1]) + 1,
            )

        end_line = region.end_line
        if end_line == 0:
            # SARIF output may omit `end_line` if equal to `start_line`, but
            # Gerrit requires it.
            end_line = region.start_line
        if region.end_column == 0:
            # end_col==0 indicates that the region goes to the end of a line,
            # but Gerrit requires that the region specify up to the start of the
            # next line.
            end_line += 1

        # SARIF uses 1-based line and column numbers, whereas Gerrit comment ranges
        # use 1-based line numbers but 0-based column numbers. Both use an exclusive
        # end column number.
        # https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#comment-range
        return {
            "start_line": region.start_line,
            "start_character": max(region.start_column - 1, 0),
            "end_line": end_line,
            "end_character": max(region.end_column - 1, 0),
        }

    def _changed_files(self, root):
        changes = self.m.buildbucket.build.input.gerrit_changes
        if not self.m.buildbucket_util.is_tryjob or not changes:
            return set()

        with self.m.context(cwd=root):
            remotes = self.m.git.get_remotes()
        remotes = [r for r in remotes if r.name == "origin"]
        assert remotes, "no remote with name 'origin'"

        remote = remotes[0].url
        gerrit_host = self.m.gerrit.host_from_remote_url(remote)
        gerrit_project = self.m.gerrit.project_from_remote_url(remote)
        if (changes[0].host, changes[0].project) != (gerrit_host, gerrit_project):
            # Don't consider any files to have changed if the repo being checked
            # is not the same repo that the pending change is in.
            return set()

        with self.m.context(cwd=root):
            return set(self.m.git.get_changed_files())
