# Copyright 2023 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import re

from recipe_engine import post_process
from PB.go.chromium.org.luci.buildbucket.proto import (
    project_config as project_config_pb2,
)
from PB.recipe_modules.fuchsia.validate_lucicfg.tests.full import InputProperties

DEPS = [
    "fuchsia/buildbucket_util",
    "fuchsia/git_checkout",
    "fuchsia/luci_config",
    "fuchsia/validate_lucicfg",
    "recipe_engine/properties",
    "recipe_engine/raw_io",
]

PROPERTIES = InputProperties
_REMOTE = "https://fuchsia.googlesource.com/infra/config"


def RunSteps(api, props):
    checkout_root, _ = api.git_checkout(_REMOTE)
    api.validate_lucicfg(checkout_root, props.validate_lucicfg_options)


def GenTests(api):
    def properties(**kwargs):
        return api.properties(
            InputProperties(
                validate_lucicfg_options=api.validate_lucicfg.options(**kwargs)
            )
        )

    nesting = "check for config deployment race conditions"

    yield (
        api.buildbucket_util.test("starlark", git_repo=_REMOTE)
        # Cover the logic for installing lucicfg from a checked-in CIPD ensure
        # file.
        + properties(lucicfg_ensure_file="cipd.ensure")
        # Cover the logic that reads builder names from cr-buildbucket.cfg.
        + api.luci_config.mock_local_config(
            "fuchsia",
            "cr-buildbucket.cfg",
            project_config_pb2.BuildbucketCfg(
                buckets=[
                    dict(
                        name="try",
                        swarming=dict(builders=[dict(name="core.x64-asan")]),
                    )
                ]
            ),
            nesting=nesting,
        )
    )

    # The second file should be validated even if the first one fails.
    yield (
        api.buildbucket_util.test(
            "first_file_invalid", tryjob=True, git_repo=_REMOTE, status="FAILURE"
        )
        + properties()
        + api.step_data(
            "validate.main.star",
            retcode=1,
            stderr=api.raw_io.output_text("error!"),
        )
        + api.post_process(post_process.MustRun, "validate.dev.star")
        # Make sure we patch in the triggering change, since this is a tryjob.
        + api.post_process(post_process.MustRun, "checkout.git rebase")
    )

    yield (
        api.buildbucket_util.test(
            "cq_new_builder", tryjob=True, git_repo=_REMOTE, status="FAILURE"
        )
        + properties()
        + api.luci_config.mock_local_config(
            "fuchsia",
            "commit-queue.cfg",
            api.validate_lucicfg.cq_config(["fuchsia/try/core.x64-asan"]),
            nesting=nesting,
        )
        + api.post_process(
            post_process.ResultReasonRE,
            re.escape(api.validate_lucicfg.NEW_BUILDERS_IN_CQ_MSG),
        )
    )

    yield (
        api.buildbucket_util.test(
            "deleting_cq_builder", tryjob=True, git_repo=_REMOTE, status="FAILURE"
        )
        + properties()
        + api.luci_config.mock_config(
            "fuchsia",
            "commit-queue.cfg",
            api.validate_lucicfg.cq_config(["fuchsia/try/core.x64-asan"]),
            nesting=nesting,
        )
        + api.post_process(
            post_process.ResultReasonRE,
            re.escape(api.validate_lucicfg.CQ_BUILDERS_DELETED_MSG),
        )
    )

    # It should be considered safe to delete an includable_only builder even if
    # it's still referenced in commit-queue.cfg.
    yield (
        api.buildbucket_util.test(
            "deleting_includable_only_builder", tryjob=True, git_repo=_REMOTE
        )
        + properties()
        + api.luci_config.mock_config(
            "fuchsia",
            "commit-queue.cfg",
            api.validate_lucicfg.cq_config(
                ["fuchsia/try/core.x64-asan"],
                includable_only=True,
            ),
            nesting=nesting,
        )
    )

    yield (
        api.buildbucket_util.test("non_main_branch", tryjob=True)
        + properties()
        + api.buildbucket_util.tryjob_branch("non-main-branch")
        # Race condition check should be skipped on branches other than "main".
        + api.post_process(
            post_process.DoesNotRun, "check for config deployment race conditions"
        )
    )
