# Copyright 2018 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Recipe for building some Breakpad tools."""

from PB.recipes.fuchsia.contrib.breakpad_tools import InputProperties

DEPS = [
    "fuchsia/buildbucket_util",
    "fuchsia/cipd_util",
    "fuchsia/git",
    "fuchsia/git_checkout",
    "fuchsia/linux_sdk",
    "fuchsia/platform_util",
    "fuchsia/toolchain",
    "fuchsia/zlib",
    "fuchsia/zstd",
    "recipe_engine/cipd",
    "recipe_engine/context",
    "recipe_engine/file",
    "recipe_engine/json",
    "recipe_engine/path",
    "recipe_engine/platform",
    "recipe_engine/step",
]

BREAKPAD_GIT = "https://chromium.googlesource.com/breakpad/breakpad"
DEPOT_TOOLS_GIT = "https://chromium.googlesource.com/chromium/tools/depot_tools.git"
RUSTC_DEMANGLE_GIT = (
    "https://fuchsia.googlesource.com/third_party/github.com/rust-lang/rustc-demangle"
)
RUSTC_DEMANGLE_TARGET = "x86_64-unknown-linux-gnu"

# CIPD packages required for building breakpad tools.
CIPD_DEPENDENCIES = (
    ("fuchsia/third_party/clang/${platform}", "integration"),
    ("fuchsia/third_party/make/${platform}", "latest"),
    ("fuchsia/third_party/ninja/${platform}", "integration"),
    ("fuchsia/third_party/rust/host/${platform}", "latest"),
)

PROPERTIES = InputProperties


def RunSteps(api, props):
    cipd_dir = api.path.start_dir / "cipd"
    with api.context(infra_steps=True):
        pkgs = api.cipd.EnsureFile()
        for pkg in CIPD_DEPENDENCIES:
            pkgs.add_package(*pkg)

        # Add the Rust target package based on the current host
        host_platform = api.cipd_util.platform_name
        host_triple = api.toolchain.PLATFORM_TO_TRIPLE[host_platform].replace(
            "arm64-apple-darwin", "aarch64-apple-darwin"
        )
        pkgs.add_package(f"fuchsia/third_party/rust/target/{host_triple}", "latest")

        api.cipd.ensure(cipd_dir, pkgs)

    # Force-install the linux sysroot so the installation step doesn't
    # pollute the top level.
    with api.linux_sdk():
        # The context API interprets bare percent signs as formatting
        # directives so they must be escaped.
        sysroot_dir = str(api.linux_sdk.sysroot).replace("%", "%%")

    rustc_demangle_src, _ = api.git_checkout(
        RUSTC_DEMANGLE_GIT, revision="0.1.21", step_name="checkout rustc_demangle"
    )

    env = {
        "CC": f"{str(cipd_dir.joinpath('bin', 'clang'))}",
        "CXX": f"{str(cipd_dir.joinpath('bin', 'clang++'))}",
        "AR": f"{str(cipd_dir.joinpath('bin', 'llvm-ar'))}",
        "LINKER": f"{str(cipd_dir.joinpath('bin', 'ld.lld'))}",
        "RANLIB": f"{str(cipd_dir.joinpath('bin', 'llvm-ranlib'))}",
        "RUSTC": f"{str(cipd_dir.joinpath('bin', 'rustc'))}",
        "CFLAGS": f"--sysroot={sysroot_dir}",
        "CXXFLAGS": f"--sysroot={sysroot_dir}",
        "LDFLAGS": f"--sysroot={sysroot_dir}",
        "CARGO_TARGET_X86_64_UNKNOWN_LINUX_GNU_RUSTFLAGS": f"-Clink-arg=--sysroot={sysroot_dir}",
    }
    with api.context(env=env, env_prefixes={"PATH": [cipd_dir / "bin"]}):
        build_rustc_demangle(api, rustc_demangle_src, RUSTC_DEMANGLE_TARGET)
        rustc_demangle_include = rustc_demangle_src / "crates" / "capi" / "include"
        rustc_demangle_lib = rustc_demangle_src.joinpath(
            "target", RUSTC_DEMANGLE_TARGET, "release"
        )

        with api.context(infra_steps=True):
            breakpad_url = props.url or BREAKPAD_GIT
            breakpad_ref = props.ref or "refs/heads/main"

            breakpad_src_dir, breakpad_revision = checkout_breakpad(
                api, breakpad_url, breakpad_ref
            )

        with api.step.nest("zlib"):
            zlib_path = api.zlib.build(
                platform=api.platform_util.host,
                install_dir=api.path.start_dir / "zlib",
            )
            zlib_include = zlib_path / "usr" / "local" / "include"
            zlib_lib = zlib_path / "usr" / "local" / "lib"
            # We remove the shared libraries because there is no way to not
            # build them, or to modify the breakpad tools build to prefer the
            # static library.
            api.file.rmglob("remove libz.so*", zlib_lib, "libz.so*")

        with api.step.nest("zstd"):
            zstd_path = api.path.start_dir / "zstd"
            api.zstd.build(
                cmake_extra_args=[
                    "-DCMAKE_INSTALL_PREFIX=%s" % zstd_path,
                    "-DZSTD_BUILD_SHARED=OFF",
                ],
            )
            zstd_include = zstd_path / "include"
            zstd_lib = zstd_path / "lib"

    # Build the two required tools, dump_syms and sym_upload using configure and
    # `make`. `make` targets must be relative paths.
    build_dir_relpath = ["src", "tools", api.platform.name]
    dump_syms_relpath = build_dir_relpath + ["dump_syms", "dump_syms"]
    build_dir = breakpad_src_dir.joinpath(*build_dir_relpath)
    dump_syms_file = breakpad_src_dir.joinpath(*dump_syms_relpath)
    with api.context(cwd=breakpad_src_dir, env_prefixes={"PATH": [cipd_dir / "bin"]}):
        # The --with-rustc-demangle flag in the breakpad tools build makes
        # assumptions about where the lib directory is, as well as prefers
        # dynamic linking so we just drive it directly here.
        configure_flags = [
            f"--sysroot={sysroot_dir}",
            "-O3",
            f"-I{str(zlib_include)}",
            f"-I{str(zstd_include)}",
            f"-I{str(rustc_demangle_include)}",
        ]
        additional_defines = [
            "-DHAVE_RUSTC_DEMANGLE",
            # TODO(fxbug.dev/64200): These defines should come from `elf.h` however the
            # Fuchsia linux sysroot is sufficiently old that it does not contain these.
            "-DEM_RISCV=243",
            "-DSHF_COMPRESSED=2048",
            "-DELFCOMPRESS_ZLIB=1",
            "-DELFCOMPRESS_ZSTD=2",
        ]
        additional_ldflags = [
            f"-L{str(rustc_demangle_lib)}",
            f"-L{str(zlib_lib)}",
            f"-L{str(zstd_lib)}",
            "-l:librustc_demangle.a",
            "-lpthread",
            "-ldl",
        ]
        configure_variables = {
            "CC": "clang",
            "CXX": "clang++",
            "CFLAGS": " ".join(configure_flags + additional_defines),
            "CPPFLAGS": " ".join(configure_flags + additional_defines),
            "CXXFLAGS": " ".join(configure_flags + additional_defines),
            "LDFLAGS": " ".join(configure_flags + additional_ldflags),
            "AR": "llvm-ar",
            "RANLIB": "llvm-ranlib",
            "NM": "llvm-nm",
            "STRIP": "llvm-strip",
            "OBJCOPY": "llvm-objcopy",
        }
        api.step(
            "configure",
            [
                "./configure",
                "--disable-dependency-tracking",
                "--disable-shared",
                "--enable-all-static",
                "--enable-zstd",
            ]
            + sorted([f"{k}={v}" for k, v in configure_variables.items()]),
        )
        # make targets must be relative paths.
        api.step("build dump_syms", ["make", api.path.join(*dump_syms_relpath)])

        # Create a cipd package definition and then register the package.
        if not api.buildbucket_util.is_dev_or_try:
            api.cipd_util.upload_package(
                "fuchsia/tools/breakpad/${platform}",
                build_dir,
                pkg_paths=[dump_syms_file],
                search_tag={"git_revision": breakpad_revision},
                repository=breakpad_url,
            )


def build_rustc_demangle(api, src, target):
    with api.step.nest("build rustc_demangle"), api.context(cwd=src):
        cargo_args = [
            "--verbose",
            f"--target={target}",
            "--release",
            "-p=rustc-demangle-capi",
        ]
        api.step("cargo build", ["cargo", "build"] + cargo_args)


def checkout_breakpad(api, url, ref):
    revision = api.git.get_remote_branch_head(url, ref)

    depot_tools_path, _ = api.git_checkout(
        DEPOT_TOOLS_GIT, step_name="checkout breakpad"
    )

    # Use gclient to fetch the DEPS.
    tree_root_dir = api.path.start_dir / "breakpad"

    api.file.ensure_directory("makedirs breakpad", tree_root_dir)

    with api.context(cwd=tree_root_dir, env_prefixes={"PATH": [depot_tools_path]}):
        # The gclient root is breakpad/.
        api.step(
            "gclient config",
            ["gclient", "config", "--name=src", "--unmanaged", "-v", url],
        )

        # This first sync seems redundant, but is necessary on a clean tree,
        # otherwise the root git repo won't yet have been pulled, so there won't be
        # anything for git to pin.
        api.step("gclient sync", ["gclient", "sync"])

        # The code is pulled by gclient sync inside breakpad/src, which is then
        # pinned to the specified revision.
        src_dir = tree_root_dir / "src"
        with api.context(cwd=src_dir):
            api.step("pin git", ["git", "checkout", revision])

            # This sync updates the dependencies to those matching |revision| as
            # specified by DEPS in the root git repo.
            step = api.step(
                "gclient sync",
                ["gclient", "sync", "-v", "--output-json", api.json.output()],
            )
            step.presentation.properties["got_revision"] = revision
        return src_dir, revision


def GenTests(api):
    yield (
        api.test("default")
        + api.platform.name("linux")
        + api.git.get_remote_branch_head(
            "git ls-remote", "9eac2058b70615519b2c4d8c6bdbfca1bd079e39"
        )
    )
