# Copyright 2019 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Recipe for rolling packages and projects to versions based on data in CIPD.

Packages that depend on opal need to get rolled based on the version
dependencies of another fuchsia prebuilt (opal). In this recipe we are
calling the prebuilt dictating the dependencies "orchestrator".

This recipe checks if there is a new version of the orchestrator.
If there is then it proceeds to roll new versions of projects and packages as
described in the rolling_packages_projects parameter.
"""

from google.protobuf import json_format as jsonpb

from PB.recipes.fuchsia.contrib.cipd_with_versions_roller import InputProperties

DEPS = [
    "fuchsia/auto_roller",
    "fuchsia/buildbucket_util",
    "fuchsia/jiri",
    "recipe_engine/buildbucket",
    "recipe_engine/cipd",
    "recipe_engine/context",
    "recipe_engine/json",
    "recipe_engine/path",
    "recipe_engine/properties",
    "recipe_engine/step",
]

PROPERTIES = InputProperties

COMMIT_MESSAGE = """[roll] Roll {roller} packages and projects

{packages}

{multiply}Test: CQ"""


def RunSteps(api, props):
    with api.context(infra_steps=True):
        if props.owners:
            with api.step.nest("owners") as presentation:
                presentation.step_summary_text = ", ".join(props.owners)

        # Import manifests
        with api.step.nest("jiri-manifests") as presentation:
            api.jiri.init(use_lock_file=True)
            presentation.logs["manifests"] = api.json.dumps(
                [
                    jsonpb.MessageToDict(m, preserving_proto_field_name=True)
                    for m in props.manifests
                ]
            )
            for manifest_config in props.manifests:
                api.jiri.import_manifest(
                    manifest=manifest_config.manifest,
                    remote=manifest_config.remote,
                    name=manifest_config.project,
                )
            api.jiri.update(run_hooks=False)
            with api.context(cwd=api.path.start_dir):
                api.jiri.run_hooks()

    deps_info = api.jiri.package([props.orchestrator_package])
    cipd_description = api.cipd.describe(
        props.orchestrator_package, props.orchestrator_ref or "latest"
    )
    orchestrator_current_version = deps_info[0]["version"]
    # Is there a new version to roll?
    orchestrator_expected_version = cipd_description.tags[0].tag
    if orchestrator_expected_version == orchestrator_current_version:
        msg = f"{props.orchestrator_package} in tree version is [{orchestrator_current_version}] and latest from cipd is [{orchestrator_expected_version}]"
        api.step.empty("manifest up-to-date; nothing to roll", step_text=msg)
        return api.auto_roller.nothing_to_roll()

    # Start rolling orchestrator
    props.rolling_packages_projects.append(
        InputProperties.ToRoll(
            type="package",
            name=props.orchestrator_package,
            version_tag=props.orchestrator_package,
            import_in=props.orchestrator_import_in,
        )
    )
    orchestrator_version_dict = {}
    orchestrator_version_dict[props.orchestrator_package] = (
        orchestrator_expected_version
    )
    # Roll all the dependencies
    package_msgs = []
    package_msg_template = "%s %s rolled from %s to %s"
    with api.step.nest("edit jiri manifests"):
        for package_project in props.rolling_packages_projects:
            package_msg = ""
            expected_version = orchestrator_version_dict[props.orchestrator_package]
            changes = api.jiri.edit_manifest(
                package_project.import_in,
                packages=[(package_project.name, expected_version)],
                name=package_project.name,
            )
            # Sometimes only a subset of packages are updated. In those cases
            # the changes list is empty.
            if changes["packages"]:
                old_version = changes["packages"][0]["old_version"]
                package_msg = package_msg_template % (
                    "package",
                    package_project.name,
                    old_version,
                    expected_version,
                )
            if package_msg:
                package_msgs.append(package_msg)

    # Update lock files
    with api.step.nest("update jiri lock files"):
        for lock in props.locks:
            api.jiri.resolve(
                local_manifest=True, output=lock.file, manifests=[lock.manifest]
            )

    roller_name = api.buildbucket.builder_name.removesuffix("-roller")

    multiply = ""
    if props.test_multipliers:
        encoded = [
            jsonpb.MessageToDict(tm, preserving_proto_field_name=True)
            for tm in props.test_multipliers
        ]
        multiply = f"\nMULTIPLY: `{api.json.dumps(encoded, indent=2)}`\n"

    message = COMMIT_MESSAGE.format(
        roller=roller_name,
        packages="\n\n".join(package_msgs),
        builder=api.buildbucket.builder_name,
        build_id=api.buildbucket_util.id,
        multiply=multiply,
    )

    # All the changes are happening inside one of the git cloned projects.
    # The "project" property points to the location of the project containing
    # the changes and we need to move to that directory for the roller to pick
    # up the changes.
    project_dir = api.path.start_dir / props.project
    with api.context(cwd=project_dir):
        change = api.auto_roller.attempt_roll(
            api.auto_roller.Options(
                remote=props.manifests[0].remote,
                dry_run=props.dry_run,
                roller_owners=props.owners,
            ),
            repo_dir=project_dir,
            commit_message=message,
        )
        rolled = change and change.success

    return api.auto_roller.raw_result(change)


def GenTests(api):
    """Tests for cipd with dependents roller."""

    def properties(**kwargs):
        props = {
            "project": "integration",
            "manifests": [
                {
                    "project": "integration",
                    "manifest": "other/dependency",
                    "remote": "sso://fuchsia/integration",
                },
                {
                    "project": "integration",
                    "manifest": "fuchsia/flower",
                    "remote": "sso://fuchsia/integration",
                },
            ],
            "locks": [{"manifest": "a/b", "file": "a/b/jiri.lock"}],
            "orchestrator_package": "fuchsia/orchestrator",
            "orchestrator_import_in": "fuchsia/prebuilts",
            "rolling_packages_projects": [
                {
                    "name": "foo",
                    "import_in": "fuchsia/prebuilts",
                },
                {
                    "name": "bar",
                    "import_in": "fuchsia/prebuilts",
                },
                {
                    "name": "baz-sdk/${platform}",
                    "import_in": "integration/fuchsia/prebuilts",
                },
            ],
            "owners": ["abc@gmail.com"],
            "dry_run": True,
            "debug_symbol_packages": ["foo/fuchsia-debug-symbols-x64"],
            "debug_symbol_gcs_buckets": ["fuchsia-debug-symbols-shortlived"],
        }
        props.update(kwargs)
        return api.properties(**props)

    package_match_test_data = api.step_data(
        "cipd describe fuchsia/orchestrator",
        api.cipd.example_describe(
            package_name="fuchsia/orchestrator",
            version="version_abc",
            test_data_tags=["git_revision:revision_abc"],
        ),
    )
    package_no_match_test_data = api.step_data(
        "cipd describe fuchsia/orchestrator",
        api.cipd.example_describe(
            package_name="fuchsia/orchestrator",
            version="version_abc",
            test_data_tags=["git_revision:revision_jkl"],
        ),
    )
    jiri_package_test_data = [
        {
            "name": "fuchsia/orchestrator",
            "path": "fuchsia/orchestrator",
            "version": "git_revision:revision_abc",
            "manifest": "manifest1",
        },
    ]

    # Version of the cipd orchestrator package is the same as the one
    # in the tree.
    yield (
        api.buildbucket_util.test("nothing_to_roll")
        + properties()
        + package_match_test_data
        + api.step_data("jiri package", api.jiri.package(jiri_package_test_data))
    )

    # Validate the build and upload path.
    yield (
        api.buildbucket_util.test(
            "orchestrator_dep_version_different_from_tree",
            builder="orchestrator-dependents-roller",
        )
        + properties(dry_run=False)
        + package_no_match_test_data
        + api.step_data("jiri package", api.jiri.package(jiri_package_test_data))
        + api.auto_roller.success()
    )

    # Validate that test multiplier works.
    yield (
        api.buildbucket_util.test(
            "rolling_packages_multiplied", builder="orchestrator-dependents-roller"
        )
        + properties(
            dry_run=False,
            test_multipliers=[{"name": "test1", "total_runs": 5}],
        )
        + package_no_match_test_data
        + api.step_data("jiri package", api.jiri.package(jiri_package_test_data))
        + api.auto_roller.success()
    )
