# Copyright 2019 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Recipe for building CMake."""

from PB.go.chromium.org.luci.common.proto.srcman.manifest import Manifest
from PB.recipes.fuchsia.contrib.cmake import InputProperties

import re

DEPS = [
    "fuchsia/buildbucket_util",
    "fuchsia/cas_util",
    "fuchsia/cipd_util",
    "fuchsia/cmake",
    "fuchsia/git_checkout",
    "fuchsia/go",
    "fuchsia/linux_sdk",
    "fuchsia/macos_sdk",
    "fuchsia/ninja",
    "fuchsia/platform_util",
    "fuchsia/windows_sdk",
    "recipe_engine/buildbucket",
    "recipe_engine/cipd",
    "recipe_engine/context",
    "recipe_engine/file",
    "recipe_engine/path",
    "recipe_engine/platform",
    "recipe_engine/properties",
    "recipe_engine/step",
]

PROPERTIES = InputProperties

BORINGSSL_GIT_URL = "https://boringssl.googlesource.com/boringssl"

GIT_URL = "https://fuchsia.googlesource.com/third_party/github.com/Kitware/CMake"

CIPD_SERVER_HOST = "chrome-infra-packages.appspot.com"


def RunSteps(api, props):
    host_platform = api.platform_util.host
    target_platform = (
        api.platform_util.platform(props.platform) if props.platform else host_platform
    )

    manifest = Manifest()

    with api.step.nest("ensure packages"), api.context(infra_steps=True):
        cipd_dir = api.path.start_dir / "cipd"
        pkgs = api.cipd.EnsureFile()
        pkgs.add_package("fuchsia/third_party/clang/${platform}", "integration")
        if target_platform.is_win:
            pkgs.add_package(
                "fuchsia/third_party/yasm/${platform}",
                "git_revision:51af4082cc898b122b88f11fd34033fc00fad81e",
            )
        ensured = api.cipd.ensure(cipd_dir, pkgs)
        for subdir, pins in ensured.items():
            directory = manifest.directories[relpath(api, cipd_dir / subdir)]
            directory.cipd_server_host = CIPD_SERVER_HOST
            for pin in pins:
                directory.cipd_package[pin.package].instance_id = pin.instance_id

    with api.linux_sdk(), api.macos_sdk(), api.windows_sdk():
        options = []
        if target_platform.is_win:
            options.extend(
                [
                    f"-DCMAKE_ASM_COMPILER={cipd_dir.joinpath('bin', 'clang-cl.exe')}",
                    f"-DCMAKE_C_COMPILER={cipd_dir.joinpath('bin', 'clang-cl.exe')}",
                    f"-DCMAKE_CXX_COMPILER={cipd_dir.joinpath('bin', 'clang-cl.exe')}",
                    f"-DCMAKE_MAKE_PROGRAM={api.ninja.path}",
                    f"-DCMAKE_SYSROOT={api.windows_sdk.sdk_dir}",
                    "-DCMAKE_MT=mt.exe",
                    "-DCMAKE_RC=rc.exe",
                ]
            )
        else:
            options.extend(
                [
                    f"-DCMAKE_ASM_COMPILER={cipd_dir.joinpath('bin', 'clang')}",
                    f"-DCMAKE_C_COMPILER={cipd_dir.joinpath('bin', 'clang')}",
                    f"-DCMAKE_CXX_COMPILER={cipd_dir.joinpath('bin', 'clang++')}",
                    f"-DCMAKE_MAKE_PROGRAM={api.ninja.path}",
                ]
            )
        if target_platform.is_linux:
            options.extend([f"-DCMAKE_SYSROOT={api.linux_sdk.sysroot}"])
        if target_platform.is_mac:
            options.extend(
                [f"-DCMAKE_SYSROOT={api.macos_sdk.sysroot}"]
                + [
                    f"-DCMAKE_{mode}_LINKER_FLAGS=-nostdlib++ {cipd_dir.joinpath('lib', 'libc++.a')}"
                    for mode in ["SHARED", "MODULE", "EXE"]
                ]
            )
        if target_platform != host_platform:
            options.extend(
                [
                    f"-DCMAKE_C_COMPILER_TARGET={target_platform.triple}",
                    f"-DCMAKE_CXX_COMPILER_TARGET={target_platform.triple}",
                    f"-DCMAKE_ASM_COMPILER_TARGET={target_platform.triple}",
                    f"-DCMAKE_SYSTEM_NAME={target_platform.os.replace('mac', 'darwin').title()}",
                    f"-DCMAKE_SYSTEM_PROCESSOR={target_platform.arch.replace('amd64', 'x86_64').replace('arm64', 'aarch64')}",
                ]
            )

        with api.step.nest("boringssl"):
            boringssl_src_dir = api.path.start_dir.joinpath("src", "boringssl")
            _, revision = api.git_checkout(
                BORINGSSL_GIT_URL,
                path=boringssl_src_dir,
                # TODO(fxbug.dev/128281): Pin boringssl to the same revision as
                # we did in clang_toolchain recipe.
                revision="01d195bd03bfff54dc99c0df0858197c71d35417",
            )
            git_checkout = manifest.directories[
                relpath(api, boringssl_src_dir)
            ].git_checkout
            git_checkout.repo_url = BORINGSSL_GIT_URL
            git_checkout.revision = revision

            boringssl_build_dir = api.path.start_dir.joinpath("src", "boringssl-build")
            boringssl_install_dir = api.path.start_dir / "boringssl"
            api.cmake.build_with_ninja(
                src_dir=boringssl_src_dir,
                build_dir=boringssl_build_dir,
                install_dir=boringssl_install_dir,
                build_type="Release",
                extra_args=[
                    "-DCMAKE_INSTALL_PREFIX=",
                    f"-DGO_EXECUTABLE={api.go.go_root.joinpath('bin', 'go')}",
                ]
                + (
                    [
                        f"-DCMAKE_ASM_NASM_COMPILER={cipd_dir.joinpath('bin', 'yasm.exe')}",
                    ]
                    if target_platform.is_win
                    else []
                )
                + options,
            )

        src_dir = api.path.start_dir.joinpath("src", "cmake")
        _, revision = api.git_checkout(
            GIT_URL, path=src_dir, fallback_ref="refs/heads/master"
        )
        git_checkout = manifest.directories[relpath(api, src_dir)].git_checkout
        git_checkout.repo_url = GIT_URL
        git_checkout.revision = revision

        build_dir = api.path.start_dir.joinpath("src", "cmake-build")
        install_dir = api.path.start_dir / "cmake"
        api.cmake.build_with_ninja(
            src_dir=src_dir,
            build_dir=build_dir,
            install_dir=install_dir,
            install_cmd="install" if api.platform.is_win else "install/strip",
            build_type="Release",
            extra_args=[
                "-DCMAKE_INSTALL_PREFIX=",
                "-DCMAKE_FIND_PACKAGE_PREFER_CONFIG=ON",
                "-DCMAKE_USE_OPENSSL=ON",
                f"-DOpenSSL_DIR={boringssl_install_dir.joinpath('lib', 'cmake', 'OpenSSL')}",
                "-DCMake_BUILD_LTO=ON",
                # TODO(phosek): build ncurses?
                "-DBUILD_CursesDialog=OFF",
            ]
            + options,
        )
        version_data = api.file.read_text(
            "read cmake revision",
            build_dir.joinpath("Source", "cmVersionConfig.h"),
            test_data='\n#define CMake_VERSION "3.22.20211015-gc5ae200"\n',
        )
        regex = re.compile(r"CMake\_VERSION (\"[a-z0-9.-]+\")")
        match = regex.search(version_data)
        if not match:
            raise api.step.StepFailure("revision string not found")  # pragma: no cover
        cmake_version = match.group(1).strip(' "')
        api.step.empty("cmake version", step_text=cmake_version)
        # TODO: CMake tests are failing, disable them until we figure out why.
        # api.ninja("test", ["test"])

    api.file.write_proto(
        "source manifest",
        install_dir / "source_manifest.json",
        manifest,
        "JSONPB",
    )

    # Upload the installation to CAS.
    api.cas_util.upload(install_dir, output_property="isolated")

    if api.buildbucket.build.builder.bucket == "prod":
        # Upload the installation to CIPD for production builds.
        api.cipd_util.upload_package(
            f"fuchsia/third_party/cmake/{target_platform}",
            install_dir,
            search_tag={"git_revision": revision},
            repository=GIT_URL,
            metadata=[("version", cmake_version)],
        )


def relpath(api, path):
    return str(api.path.relpath(path, api.path.start_dir))


def GenTests(api):
    for platform in ("linux", "mac", "win"):
        yield (
            api.buildbucket_util.test(
                platform, bucket="prod", git_repo=GIT_URL, revision="a" * 40
            )
            + api.platform.name(platform)
        )

    yield (
        api.buildbucket_util.test(
            "cross", bucket="prod", git_repo=GIT_URL, revision="a" * 40
        )
        + api.platform.name("mac")
        + api.properties(platform="mac-arm64")
    )
