# Copyright 2018 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Recipe for building Dart toolchain."""

from recipe_engine.config import Enum
from recipe_engine.recipe_api import Property

DEPS = [
    "fuchsia/cipd_util",
    "fuchsia/git_checkout",
    "fuchsia/goma",
    "fuchsia/linux_sdk",
    "fuchsia/macos_sdk",
    "fuchsia/ninja",
    "recipe_engine/buildbucket",
    "recipe_engine/context",
    "recipe_engine/file",
    "recipe_engine/json",
    "recipe_engine/path",
    "recipe_engine/platform",
    "recipe_engine/properties",
    "recipe_engine/raw_io",
    "recipe_engine/step",
]

DEPOT_TOOLS_GIT = "https://chromium.googlesource.com/chromium/tools/depot_tools"
DART_SDK_GIT = "https://dart.googlesource.com/sdk.git"

PROPERTIES = {
    "repository": Property(kind=str, help="Git repository URL", default=DART_SDK_GIT),
    "revision": Property(kind=str, help="Revision", default=None),
    "host_cpu": Property(
        kind=Enum("arm64", "x64"),
        help="GN $host_cpu toolchain will run on",
        default=None,
    ),
    "host_os": Property(
        kind=Enum("linux", "mac"),
        help="GN $host_os toolchain will run on",
        default=None,
    ),
}

# Fuchsia targets the toolchain must support.
FUCHSIA_TARGETS = [
    "arm64",
    "x64",
    "riscv64",
]

GN_CIPD_CPU_MAP = {
    "x64": "amd64",
}


def RunSteps(api, repository, revision, host_cpu, host_os):
    gitiles_commit = api.buildbucket.build.input.gitiles_commit
    if gitiles_commit.host and gitiles_commit.project:
        repository = f"https://{gitiles_commit.host}/{gitiles_commit.project}"

    if not host_os:
        host_os = api.platform.name
    cipd_platform = f"{host_os}-{GN_CIPD_CPU_MAP.get(host_cpu, host_cpu)}"

    ninja_jobs = api.goma.jobs

    # TODO(phosek): Move this to a named cache so it's shared across builds.
    depot_tools_dir, _ = api.git_checkout(
        DEPOT_TOOLS_GIT, submodules=False, step_name="depot_tools"
    )

    dart_dir = api.path.start_dir / "dart"
    sdk_dir = dart_dir / "sdk"

    _, revision = api.git_checkout(
        repository, path=sdk_dir, submodules=False, step_name="dart-sdk"
    )

    # Use gclient to fetch the DEPS, but don't let it change dart/sdk itself.
    with api.context(
        infra_steps=True, cwd=dart_dir, env_prefixes={"PATH": [depot_tools_dir]}
    ):
        api.step(
            "gclient config",
            [
                "gclient",
                "config",
                "--unmanaged",
                # TODO(mcgrathr): Set --cache-dir?
                "-v",
                DART_SDK_GIT,
            ],
        )
        api.step(
            "gclient sync",
            [
                "gclient",
                "sync",
                "--no-history",
                "-v",
                "--output-json",
                api.json.output(),
            ],
        )

    dart_targets = sorted(
        set(
            [host_cpu]
            + [
                "sim" + cpu + "_" + host_cpu if cpu != host_cpu else cpu
                for cpu in FUCHSIA_TARGETS
            ]
        )
    )

    # build_cpu -> where gen_snapshot will be built
    # host_cpu -> where gen_snapshot will run
    # cpu -> where gen_snapshot will target
    build_cpu = "x64"

    def out_dir(cpu):
        return sdk_dir.joinpath(
            "xcodebuild" if api.platform.is_mac else "out",
            "Release"
            + ("X" if host_cpu == cpu and host_cpu != build_cpu else "")
            + cpu.upper(),
        )

    with api.linux_sdk(), api.macos_sdk(), api.goma(), api.context(cwd=sdk_dir):
        # TODO(mcgrathr): Move this logic into a host_build recipe module shared
        # with gcc_toolchain.py and others.
        if host_os == "linux":
            host_sysroot = api.linux_sdk.sysroot
        elif host_os == "mac":
            host_sysroot = api.macos_sdk.sysroot
        else:  # pragma: no cover
            assert False, "what platform?"

        gn_common = [
            "tools/gn.py",
            "-v",
            "--mode=release",
            "--goma",
            # It actually wants a sysroot for the build host regardless of what
            # Dart target CPU the build is configured for, but wants it in the
            # form `%(target_cpu)s=%(host_sysroot)s`.
            "--target-sysroot="
            + ",".join(f"{cpu}={host_sysroot}" for cpu in dart_targets),
        ]

        api.step(
            f"gn host ({host_cpu})",
            gn_common + [f"--arch={host_cpu}", "--platform-sdk"],
        )
        for cpu in dart_targets:
            if cpu != host_cpu:
                api.step(f"gn {cpu}", gn_common + [f"--arch={cpu}"])
        api.ninja(
            f"build host ({host_cpu}) SDK",
            [
                "create_sdk",
                "gen_snapshot",
                "gen_snapshot_product",
            ],
            build_dir=out_dir(host_cpu),
            ninja_jobs=ninja_jobs,
        )
        for cpu in dart_targets:
            api.ninja(
                f"build {cpu}",
                [
                    "gen_snapshot_fuchsia",
                    "gen_snapshot_product_fuchsia",
                ],
                build_dir=out_dir(cpu),
                ninja_jobs=ninja_jobs,
            )

    with api.step.nest("install"):
        pkg_dir = api.path.cleanup_dir / "dart-sdk"
        api.file.copytree(
            "install dart-sdk",
            out_dir(host_cpu) / "dart-sdk",
            pkg_dir,
            symlinks=True,
        )
        for cpu, is_host in [(host_cpu, True)] + [(cpu, False) for cpu in dart_targets]:
            exe_dir = out_dir(cpu) / "exe.stripped"
            target_cpu = cpu.split("_")[0]
            if target_cpu.startswith("sim"):
                target_cpu = target_cpu[3:]
            for tool in ["gen_snapshot", "gen_snapshot_product"]:
                src = exe_dir.joinpath(tool + ("" if is_host else "_fuchsia"))
                dst_name = f"{tool}.{host_os if is_host else 'fuchsia'}-{target_cpu}"
                dst = pkg_dir.joinpath("bin", dst_name)
                api.file.copy(f"install {dst_name}", src, dst)

    dart_sdk_version = api.file.read_text(
        "read dart-sdk version",
        pkg_dir / "version",
        test_data="2.0.0-edge." + revision + "\n",
    ).strip()

    api.cipd_util.upload_package(
        "fuchsia/dart-sdk/" + cipd_platform,
        pkg_dir,
        search_tag={"git_revision": revision},
        repository=repository,
        metadata=[("dart_sdk_version", dart_sdk_version)],
    )


def GenTests(api):
    revision = "a" * 40
    for platform in ("linux", "mac"):
        for arch in ("x64", "arm64"):
            yield (
                api.test(platform + "-" + arch)
                + api.properties(host_cpu=arch)
                + api.platform.name(platform)
                + api.step_data(
                    "dart-sdk.git rev-parse", api.raw_io.stream_output_text(revision)
                )
                + api.step_data("gclient sync", api.json.output({"solutions": {}}))
                + api.buildbucket.ci_build(git_repo=DART_SDK_GIT, revision=revision)
            )
