# Copyright 2019 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Recipe for building and publishing fastboot to CIPD."""

import re

DEPS = [
    "fuchsia/buildbucket_util",
    "fuchsia/cas_util",
    "fuchsia/cipd_util",
    "fuchsia/git",
    "fuchsia/git_checkout",
    "fuchsia/macos_sdk",
    "fuchsia/ninja",
    "fuchsia/platform_util",
    "fuchsia/toolchain",
    "fuchsia/zlib",
    "recipe_engine/buildbucket",
    "recipe_engine/cipd",
    "recipe_engine/context",
    "recipe_engine/file",
    "recipe_engine/path",
    "recipe_engine/platform",
    "recipe_engine/step",
]

# List of checkout paths under the Android root that we actually need to build
# fastboot.  We do this rather than use `repo` because the android manifest is
# something like 1000 repos and we only need something like 10, so this saves a
# huge amount of time, network bandwidth, and I/O during checkout.
REPO_PATHS = [
    ("build", "soong"),
    ("development",),
    ("external", "avb"),
    ("external", "boringssl"),
    ("external", "fmtlib"),
    ("external", "googletest"),
    ("system", "core"),
    ("system", "extras"),
    ("system", "libbase"),
    ("system", "libziparchive"),
    ("system", "logging"),
    ("system", "tools", "mkbootimg"),
]


ANDROID_SOURCE_BRANCH = "main"
FASTBOOT_CIPD_PKG = "fuchsia/third_party/fastboot/${platform}"


def RunSteps(api):
    with api.step.nest("ensure packages"), api.context(infra_steps=True):
        cipd_dir = api.path.start_dir / "cipd"
        pkgs = api.cipd.EnsureFile()
        pkgs.add_package("fuchsia/third_party/clang/${platform}", "integration")
        if api.platform.is_linux:
            pkgs.add_package(
                "fuchsia/third_party/sysroot/linux",
                "integration",
                "sysroot",
            )
        api.cipd.ensure(cipd_dir, pkgs)

    src_dir = api.path.start_dir.joinpath("android", "src")
    api.file.ensure_directory("ensure source directory", src_dir)

    with api.step.nest("clone android repos"), api.context(infra_steps=True):
        with api.context(cwd=src_dir):
            for repo_path in REPO_PATHS:
                repo_name = "/".join(repo_path)
                repo_url = repo_url_for(repo_name)
                repo_full_path = src_dir.joinpath(*repo_path)
                api.git_checkout(
                    repo_url,
                    path=repo_full_path,
                    fallback_ref=ANDROID_SOURCE_BRANCH,
                    step_name=f"checkout {repo_name}",
                )

    # Future work: consider allowing specifying target platform via props, to
    # allow cross-compiling e.g. for mac-x64 from mac-arm64
    target_platform = api.platform_util.host

    # Pairs of:
    # 1. source path under boringssl root (itself external/boringssl
    #    under the android source root)
    # 2. object file name
    boringssl_objs = [
        (["src", "crypto", "crypto.c"], "crypto.o"),
        (["src", "crypto", "err", "err.c"], "err.o"),
        (["err_data.c"], "err_data.o"),
        (["src", "crypto", "fipsmodule", "rand", "fork_detect.c"], "fork_detect.o"),
        (["src", "crypto", "mem.c"], "mem.o"),
        (["src", "crypto", "fipsmodule", "sha", "sha256.c"], "sha256.o"),
        (["src", "crypto", "thread_pthread.c"], "thread_pthread.o"),
        (["src", "crypto", "fipsmodule", "rand", "urandom.c"], "urandom.o"),
    ]

    arch_specific_boringssl_objs = {
        "linux-arm64": [
            (["src", "crypto", "cpu_aarch64_linux.c"], "cpu_aarch64_linux.o"),
            (
                ["linux-aarch64", "crypto", "fipsmodule", "sha256-armv8-linux.S"],
                "sha256-armv8-linux.o",
            ),
        ],
        "linux-amd64": [
            (["src", "crypto", "cpu_intel.c"], "cpu_intel.o"),
            (
                ["linux-x86_64", "crypto", "fipsmodule", "sha256-x86_64-linux.S"],
                "sha256-x86_64-linux.o",
            ),
        ],
        "mac-arm64": [
            (["src", "crypto", "cpu_aarch64_apple.c"], "cpu_aarch64_apple.o"),
            (
                ["apple-aarch64", "crypto", "fipsmodule", "sha256-armv8-apple.S"],
                "sha256-armv8-apple.o",
            ),
        ],
        "mac-amd64": [
            (["src", "crypto", "cpu_intel.c"], "cpu_intel.o"),
            (
                ["apple-x86_64", "crypto", "fipsmodule", "sha256-x86_64-apple.S"],
                "sha256-x86_64-apple.o",
            ),
        ],
    }
    boringssl_objs.extend(arch_specific_boringssl_objs[target_platform.platform])

    # As with boringssl_objs, but relative to src_dir
    fastboot_objs = [
        (
            ["build", "soong", "cc", "libbuildversion", "libbuildversion.cpp"],
            "libbuildversion.o",
        ),
        (["external", "fmtlib", "src", "format.cc"], "format.o"),
        (["system", "core", "diagnose_usb", "diagnose_usb.cpp"], "diagnose_usb.o"),
        (["system", "core", "fastboot", "bootimg_utils.cpp"], "bootimg_utils.o"),
        (["system", "core", "fastboot", "fastboot.cpp"], "fastboot.o"),
        (["system", "core", "fastboot", "fastboot_driver.cpp"], "fastboot_driver.o"),
        (["system", "core", "fastboot", "filesystem.cpp"], "filesystem.o"),
        (["system", "core", "fastboot", "fs.cpp"], "fs.o"),
        (["system", "core", "fastboot", "main.cpp"], "main.o"),
        (["system", "core", "fastboot", "socket.cpp"], "socket.o"),
        (["system", "core", "fastboot", "storage.cpp"], "storage.o"),
        (
            ["system", "core", "fastboot", "super_flash_helper.cpp"],
            "super_flash_helper.o",
        ),
        (["system", "core", "fastboot", "task.cpp"], "task.o"),
        (["system", "core", "fastboot", "tcp.cpp"], "tcp.o"),
        (["system", "core", "fastboot", "udp.cpp"], "udp.o"),
        (["system", "core", "fastboot", "util.cpp"], "util.o"),
        (
            ["system", "core", "fastboot", "vendor_boot_img_utils.cpp"],
            "vendor_boot_img_utils.o",
        ),
        (["system", "core", "fs_mgr", "liblp", "builder.cpp"], "builder.o"),
        (
            ["system", "core", "fs_mgr", "liblp", "property_fetcher.cpp"],
            "property_fetcher.o",
        ),
        (
            ["system", "core", "fs_mgr", "liblp", "super_layout_builder.cpp"],
            "super_layout_builder.o",
        ),
        (["system", "core", "fs_mgr", "liblp", "images.cpp"], "images.o"),
        (
            ["system", "core", "fs_mgr", "liblp", "partition_opener.cpp"],
            "partition_opener.o",
        ),
        (["system", "core", "fs_mgr", "liblp", "reader.cpp"], "reader.o"),
        (["system", "core", "fs_mgr", "liblp", "utility.cpp"], "utility.o"),
        (["system", "core", "fs_mgr", "liblp", "writer.cpp"], "writer.o"),
        # if we need to support windows, move all of these with _unix to the arch_specific section
        (
            ["system", "core", "libcutils", "socket_inaddr_any_server_unix.cpp"],
            "socket_inaddr_any_server_unix.o",
        ),
        (
            ["system", "core", "libcutils", "socket_network_client_unix.cpp"],
            "socket_network_client_unix.o",
        ),
        (["system", "core", "libcutils", "sockets.cpp"], "sockets.o"),
        (["system", "core", "libcutils", "sockets_unix.cpp"], "sockets_unix.o"),
        (["system", "core", "libsparse", "backed_block.cpp"], "backed_block.o"),
        (["system", "core", "libsparse", "output_file.cpp"], "output_file.o"),
        (["system", "core", "libsparse", "sparse.cpp"], "sparse.o"),
        (["system", "core", "libsparse", "sparse_crc32.cpp"], "sparse_crc32.o"),
        (["system", "core", "libsparse", "sparse_err.cpp"], "sparse_err.o"),
        (["system", "core", "libsparse", "sparse_read.cpp"], "sparse_read.o"),
        (["system", "extras", "ext4_utils", "ext4_sb.cpp"], "ext4_sb.o"),
        (["system", "extras", "ext4_utils", "ext4_utils.cpp"], "ext4_utils.o"),
        (["system", "libbase", "errors_unix.cpp"], "errors_unix.o"),
        (["system", "libbase", "file.cpp"], "file.o"),
        (["system", "libbase", "logging.cpp"], "logging.o"),
        (["system", "libbase", "mapped_file.cpp"], "mapped_file.o"),
        (["system", "libbase", "parsebool.cpp"], "parsebool.o"),
        (["system", "libbase", "parsenetaddress.cpp"], "parsenetaddress.o"),
        (["system", "libbase", "posix_strerror_r.cpp"], "posix_strerror_r.o"),
        (
            ["system", "libbase", "properties.cpp"],
            "libbase_properties.o",
        ),  # properties.o would collide
        (["system", "libbase", "stringprintf.cpp"], "stringprintf.o"),
        (["system", "libbase", "strings.cpp"], "strings.o"),
        (["system", "libbase", "threads.cpp"], "threads.o"),
        (["system", "libziparchive", "zip_archive.cc"], "zip_archive.o"),
        (
            ["system", "libziparchive", "zip_archive_stream_entry.cc"],
            "zip_archive_stream_entry.o",
        ),
        (["system", "libziparchive", "zip_cd_entry_map.cc"], "zip_cd_entry_map.o"),
        (["system", "libziparchive", "zip_error.cpp"], "zip_error.o"),
        (["system", "logging", "liblog", "logger_write.cpp"], "logger_write.o"),
        (
            ["system", "logging", "liblog", "properties.cpp"],
            "logging_properties.o",
        ),  # properties.o would collide
    ]
    arch_specific_fastboot_objs = {
        "linux-amd64": [
            (["system", "core", "fastboot", "usb_linux.cpp"], "usb_linux.o"),
        ],
        "linux-arm64": [
            (["system", "core", "fastboot", "usb_linux.cpp"], "usb_linux.o"),
        ],
        "mac-amd64": [
            (["system", "core", "fastboot", "usb_osx.cpp"], "usb_osx.o"),
        ],
        "mac-arm64": [
            (["system", "core", "fastboot", "usb_osx.cpp"], "usb_osx.o"),
        ],
    }
    fastboot_objs.extend(arch_specific_fastboot_objs[target_platform.platform])

    fastboot_dir = src_dir.joinpath("system", "core", "fastboot")
    with api.context(cwd=fastboot_dir):
        repository = api.git.get_remote_url("origin")
        revision = api.git.get_hash()

    out_dir = api.path.mkdtemp("fastboot") / "out"
    api.file.ensure_directory("ensure output directory", out_dir)
    # Generate a suitable platform_tools_version.h in the root of out_dir
    platform_tools_template = api.file.read_text(
        "read platform tools version",
        src_dir.joinpath("development", "sdk", "plat_tools_source.prop_template"),
        test_data="Pkg.UserSrc=false\nPkg.Revision=34.0.5\n",
    )
    version_re = re.compile(r"Pkg.Revision=(\d+\.\d+\.\d+)")
    match = version_re.search(platform_tools_template)
    version = match.group(1)
    api.file.write_text(
        "write platform_tools_version.h",
        out_dir / "platform_tools_version.h",
        f'#define PLATFORM_TOOLS_VERSION "{version}"\n',
    )

    with api.macos_sdk(), api.context(cwd=src_dir):
        sysroot_dir = get_sysroot(api, cipd_dir, target_platform)
        if api.platform.is_linux:
            with api.step.nest("zlib"):
                target_triple = api.toolchain.PLATFORM_TO_TRIPLE[
                    target_platform.platform
                ]
                zlib_root = api.path.mkdtemp("zlib")
                zlib_build_dir = zlib_root / "build"
                api.file.ensure_directory("create zlib_build_dir", zlib_build_dir)
                zlib_install_dir = zlib_root / "install"
                api.file.ensure_directory("create zlib_install_dir", zlib_install_dir)
                cc = cipd_dir.joinpath("bin", "clang")
                cxx = cipd_dir.joinpath("bin", "clang++")
                ninja = api.ninja.path
                api.zlib.build(
                    cmake_extra_args=[
                        "-GNinja",
                        f"-DCMAKE_MAKE_PROGRAM={ninja}",
                        f"-DCMAKE_SYSROOT={sysroot_dir}",
                        "-DCMAKE_INSTALL_PREFIX=",
                        f"-DCMAKE_C_COMPILER={cc}",
                        f"-DCMAKE_CXX_COMPILER={cxx}",
                        f"-DCMAKE_ASM_COMPILER={cc}",
                        f"-DCMAKE_C_FLAGS=--target={target_triple}",
                        f"-DCMAKE_CXX_FLAGS=--target={target_triple}",
                        f"-DCMAKE_ASM_FLAGS=--target={target_triple}",
                    ],
                    platform=target_platform,
                    ninja_jobs=api.platform.cpu_count,
                    build_dir=zlib_build_dir,
                    install_dir=zlib_install_dir,
                )

        boringssl_include = src_dir.joinpath("external", "boringssl", "src", "include")
        cflags = [
            "-O2",
            "-c",
            "--sysroot",
            sysroot_dir,
            f"-I{boringssl_include}",
        ]

        def boringssl_build_cmd(src_elems, obj):
            return [
                cipd_dir.joinpath("bin", "clang"),
                *cflags,
                "-o",
                out_dir / obj,
                src_dir.joinpath(*src_elems),
            ]

        include_dirs = [
            src_dir.joinpath("build", "soong", "cc", "libbuildversion", "include"),
            src_dir.joinpath("external", "avb"),
            src_dir.joinpath("external", "boringssl", "src", "include"),
            src_dir.joinpath("external", "fmtlib", "include"),
            src_dir.joinpath("external", "googletest", "googletest", "include"),
            src_dir.joinpath("system", "core", "diagnose_usb", "include"),
            src_dir.joinpath("system", "core", "fs_mgr", "liblp", "include"),
            src_dir.joinpath("system", "core", "fs_mgr", "libstorage_literals"),
            src_dir.joinpath("system", "core", "libcutils", "include"),
            src_dir.joinpath("system", "core", "libsparse", "include"),
            src_dir.joinpath("system", "extras", "ext4_utils", "include"),
            src_dir.joinpath("system", "libbase", "include"),
            src_dir.joinpath("system", "libziparchive", "incfs_support", "include"),
            src_dir.joinpath("system", "libziparchive", "include"),
            src_dir.joinpath("system", "logging", "liblog", "include"),
            src_dir.joinpath("system", "tools", "mkbootimg", "include", "bootimg"),
            out_dir,  # for platform_tools_version.h
        ]
        if api.platform.is_linux:
            include_dirs.append(zlib_install_dir / "include")  # for zlib.h

        cxx_include = [f"-I{include_dir}" for include_dir in include_dirs]
        cxxflags = [
            "-O2",
            "-c",
            "--sysroot",
            sysroot_dir,
            "-std=c++20",  # required for std::span, used by libziparchive
            "-DZLIB_CONST",
            "-D_FILE_OFFSET_BITS=64",
            "-Wno-c++11-narrowing",
            "-Wno-inconsistent-missing-override",
            "-Wno-deprecated-anon-enum-enum-conversion",
            "-Wno-deprecated-declarations",
            *cxx_include,
        ]
        if api.platform.is_linux:
            cxxflags.append("-D_XOPEN_SOURCE=700")  # for POSIX but not GNU strerror_r

        def fastboot_build_cmd(src_elems, obj):
            return [
                cipd_dir.joinpath("bin", "clang++"),
                *cxxflags,
                "-o",
                out_dir / obj,
                src_dir.joinpath(*src_elems),
            ]

        def fastboot_link_cmd(objs, destfile):
            LDFLAGS = []
            if api.platform.is_mac:
                LDFLAGS.extend(
                    [
                        cipd_dir.joinpath("lib", "libc++.a"),
                        "-framework",
                        "CoreFoundation",
                        "-framework",
                        "IOKit",
                        "-lz",
                    ]
                )
            elif api.platform.is_linux:
                LDFLAGS.extend(
                    [
                        zlib_install_dir.joinpath("lib", "libz.a"),
                        "-lc++",
                    ]
                )
            return [
                cipd_dir.joinpath("bin", "clang++"),
                "-std=c++20",
                "--sysroot",
                sysroot_dir,
                "-nostdlib++",
                "-stdlib=libc++",
                "-o",
                destfile,
                *objs,
                *LDFLAGS,
            ]

        with api.step.nest("build boringssl"):
            for src_path_elems, obj_path in boringssl_objs:
                build_cmd = boringssl_build_cmd(
                    ["external", "boringssl", *src_path_elems], obj_path
                )
                api.step(f"build {obj_path}", build_cmd)

        with api.step.nest("build fastboot"):
            for src_path_elems, obj_path in fastboot_objs:
                build_cmd = fastboot_build_cmd(src_path_elems, obj_path)
                api.step(f"build {obj_path}", build_cmd)

            out_bin_dir = out_dir / "bin"
            api.file.ensure_directory("ensure fastboot bin directory", out_bin_dir)
            all_objfiles = [
                out_dir / item[1] for item in [*boringssl_objs, *fastboot_objs]
            ]
            fastboot = out_bin_dir / "fastboot"
            link_cmd = fastboot_link_cmd(all_objfiles, fastboot)
            api.step("link fastboot", link_cmd)

    upload_paths = [fastboot]
    if api.buildbucket_util.is_dev_or_try:
        api.cas_util.upload(
            out_bin_dir, upload_paths=upload_paths, output_property="isolated"
        )
    else:
        api.cipd_util.upload_package(
            FASTBOOT_CIPD_PKG,
            out_bin_dir,
            pkg_paths=upload_paths,
            search_tag={"git_revision": revision},
            metadata=[("git_branch", ANDROID_SOURCE_BRANCH)],
            repository=repository,
        )


def repo_url_for(path):
    return f"https://android.googlesource.com/platform/{path}"


def get_sysroot(api, cipd_dir, platform):
    if platform.is_linux:
        return cipd_dir / "sysroot"
    if platform.is_mac:
        return api.macos_sdk.sysroot
    raise api.step.StepFailure("unsupported platform")  # pragma: no cover


def GenTests(api):
    yield (
        api.test("mac")
        + api.platform.name("mac")
        + api.buildbucket.try_build(
            project="android",
            bucket="try",
            git_repo="https://android.googlesource.com/platform/system/core",
        )
    )
    yield (
        api.test("linux")
        + api.buildbucket.ci_build(
            project="android",
            bucket="ci",
            git_repo="https://android.googlesource.com/platform/system/core",
        )
    )
