# Copyright 2018 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Recipe for building GCC toolchain."""

import contextlib
import pipes
import re

from PB.recipes.fuchsia.contrib.gcc_toolchain import InputProperties

DEPS = [
    "fuchsia/buildbucket_util",
    "fuchsia/cas_util",
    "fuchsia/cipd_util",
    "fuchsia/git",
    "fuchsia/git_checkout",
    "fuchsia/goma",
    "fuchsia/macos_sdk",
    "fuchsia/ninja",
    "fuchsia/toolchain",
    "fuchsia/zstd",
    "recipe_engine/buildbucket",
    "recipe_engine/cipd",
    "recipe_engine/context",
    "recipe_engine/file",
    "recipe_engine/path",
    "recipe_engine/platform",
    "recipe_engine/properties",
    "recipe_engine/raw_io",
    "recipe_engine/step",
]

PROPERTIES = InputProperties

BINUTILS_GIT = "https://gnu.googlesource.com/binutils-gdb"
BINUTILS_REF = "refs/heads/master"

GCC_GIT = "https://gnu.googlesource.com/gcc"
GCC_REF = "refs/heads/releases/gcc-13"

GCC_PREREQS = {
    "gmp": "6.2.1",
    "mpfr": "4.1.0",
    "mpc": "1.2.1",
    "isl": "0.24",
}

BINUTILS_PREREQS = {
    pkg: version for pkg, version in GCC_PREREQS.items() if pkg in ["gmp", "mpfr"]
}

GCC_PATCHES = [
    "0001-libiberty-Make-safe-ctype.h-safe-in-C++.patch",
]

# Bump this number whenever changing this recipe in ways that affect the
# package built without also changing any upstream revision pin.
# TODO(crbug.com/947158): Remove this when the recipes repo/rev are available.
RECIPE_SALT = ""


def RunSteps(api, props):
    use_goma = api.platform.arch != "arm"
    # TODO(mcgrathr): temporarily disable goma while working on some bugs
    use_goma = False
    if use_goma:  # pragma: no cover
        compile_jobs = api.goma.jobs
        goma_context = api.goma()
    else:
        compile_jobs = api.platform.cpu_count
        goma_context = contextlib.nullcontext()

    prod = api.buildbucket.build.builder.bucket == "prod"
    host_platform = api.cipd_util.platform_name
    host_triple = api.toolchain.PLATFORM_TO_TRIPLE[host_platform]

    binutils_dir = api.path.start_dir / "binutils-gdb"
    gcc_dir = api.path.start_dir / "gcc"

    _, binutils_revision = api.git_checkout(
        repo=BINUTILS_GIT,
        path=binutils_dir,
        step_name="binutils-gdb",
        fallback_ref=BINUTILS_REF,
    )
    binutils_pkgversion = f"{BINUTILS_GIT} {binutils_revision}"
    _, gcc_revision = api.git_checkout(
        repo=GCC_GIT, path=gcc_dir, step_name="gcc", fallback_ref=GCC_REF
    )
    gcc_pkgversion = f"{GCC_GIT} {gcc_revision}"

    for patch in [api.resource(patch) for patch in GCC_PATCHES]:
        api.git(
            f"apply {api.path.basename(patch)}",
            "-C",
            gcc_dir,
            "apply",
            "--verbose",
            patch,
        )

    with api.step.nest("ensure source prerequisites"):
        with api.context(infra_steps=True):
            for source_dir, prereqs in [
                (gcc_dir, GCC_PREREQS),
                (binutils_dir, BINUTILS_PREREQS),
            ]:
                source_pkgs = api.cipd.EnsureFile()
                for name, version in prereqs.items():
                    source_pkgs.add_package(
                        f"fuchsia/third_party/source/{name}", f"version:{version}", name
                    )
                    api.cipd.ensure(source_dir, source_pkgs)
        # These files come out of the CIPD install with timestamps that
        # aren't necessarily newer than configure.ac and Makefile.am, and
        # the isl makefile unconditionally tries to regenerate them without
        # requiring --enable-maintainer-mode.
        for isl_file in ["aclocal.m4", "configure", "Makefile.in", "isl_config.h.in"]:
            api.step(
                f"touch isl/{isl_file}", ["touch", gcc_dir.joinpath("isl", isl_file)]
            )

    with api.step.nest("ensure_packages"):
        with api.context(infra_steps=True):
            pkgs = api.cipd.EnsureFile()
            pkgs.add_package("fuchsia/third_party/clang/${platform}", "integration")
            pkgs.add_package("fuchsia/third_party/make/${platform}", "version:4.3")
            pkgs.add_package("fuchsia/third_party/libtool/${platform}", "version:2.4.6")
            pkgs.add_package(
                "fuchsia/third_party/pkg-config/${platform}", "version:0.29.2"
            )
            pkgs.add_package("fuchsia/third_party/bison/${platform}", "version:3.7")
            pkgs.add_package("fuchsia/third_party/flex/${platform}", "version:2.6.4")
            pkgs.add_package("fuchsia/third_party/m4/${platform}", "version:1.4.18")
            pkgs.add_package("fuchsia/third_party/texinfo/${platform}", "version:6.8.0")
            if api.platform.name == "linux":
                pkgs.add_package("fuchsia/third_party/sysroot/linux", "latest")
            cipd_dir = api.path.start_dir / "cipd"
            api.cipd.ensure(cipd_dir, pkgs)

    # This is a no-op on non-Mac.  Do it just once around all the logic that
    # needs any xcode bits.  Doing it multiple times separately seems to
    # produce anomalous results.
    with api.macos_sdk():
        if api.platform.name == "linux":
            host_sysroot = cipd_dir
        elif api.platform.name == "mac":
            host_sysroot = api.macos_sdk.sysroot
            # Log ld versions because it's been a bug issue in the past.
            api.step("/usr/bin/ld -v", ["/usr/bin/ld", "-v"])
        else:  # pragma: no cover
            assert False, "what platform?"

        staging_dir = api.path.start_dir / "staging"
        pkg_name = f"gcc-{api.platform.name.replace('mac', 'darwin')}"
        pkg_dir = staging_dir / pkg_name
        api.file.ensure_directory("create pkg dir", pkg_dir)

        # Some of the makefile logic splits $CC at its first word and injects
        # a switch there.  So make $CC and $CXX be single words by writing
        # little scripts.  Autoconf does some checks with CPPFLAGS but not
        # CFLAGS and other checks with CFLAGS but not CPPFLAGS.  The sysroot
        # is necessary for all cases, so fold that into the script too so it's
        # impossible to omit it in any $CC or $CXX invocation.
        cc_path = staging_dir / "host-cc"
        cxx_path = staging_dir / "host-cxx"
        gomacc_path = api.goma.goma_dir / "gomacc" if use_goma else ""
        for script, compiler in [(cc_path, "clang"), (cxx_path, "clang++")]:
            compiler_path = cipd_dir.joinpath("bin", compiler)
            if api.platform.name == "mac" and compiler == "clang++":
                # Our host toolchain for Mac provides static libc++ but doesn't
                # know how to link it in by itself.  Things in LIBS or LDFLAGS
                # get put onto `ar` command lines for static libraries, which
                # doesn't go well.  A link input on non-linking clang++ command
                # lines is useless though harmless, but it generates a lot of
                # warning noise that makes the build logs hard to read and slow
                # to collect.  So the wrapper script tries to add it only to
                # linking command lines.
                script_text = """#!/bin/sh
extra=(-nostdlib++ %s)
for arg; do
  case "$arg" in
  -[cE])
    extra=()
    break
    ;;
  esac
done
exec %s %s -no-canonical-prefixes --sysroot=%s "$@" "${extra[@]}"
""" % (
                    cipd_dir.joinpath("lib", "libc++.a"),
                    gomacc_path,
                    compiler_path,
                    host_sysroot,
                )
            else:
                script_text = f"""#!/bin/sh
{gomacc_path} {compiler_path} -no-canonical-prefixes --sysroot={host_sysroot} "$@\"
"""
            api.file.write_text(
                f"write {api.path.basename(script)} script", script, script_text
            )
            api.step(
                f"make {api.path.basename(script)} executable",
                ["chmod", "+x", script],
            )

        host_cflags = "-O3"
        if api.platform.name != "mac":
            # LTO works for binutils on Linux but fails on macOS.
            host_cflags += " -flto"
        if api.platform.name == "linux":
            # The zstd static library needs -lpthread but that doesn't get
            # plumbed properly.  For binutils, it relies on pkg-config and that
            # fails to deliver -lpthread.  For GCC, configure ignores LIBS
            # while trying to link -lzstd with the -L implied by the
            # --with-zstd= path so there's no way to tell it.  The -pthread
            # driver switch implies -lpthread when linking and is otherwise
            # generally harmless.
            host_cflags += " -pthread"
        host_compiler_args = {
            "CC": f"{cc_path}",
            "CXX": f"{cxx_path}",
            "CFLAGS": host_cflags,
            "CXXFLAGS": host_cflags,
            "AR": cipd_dir.joinpath("bin", "llvm-ar"),
            "RANLIB": cipd_dir.joinpath("bin", "llvm-ranlib"),
            "M4": f"{cipd_dir.joinpath('bin', 'm4')}",
        }

        if api.platform.name != "mac":
            # Always link libc++ statically in case a shared library is available.
            host_compiler_args["CXXFLAGS"] += " -static-libstdc++"

        host_compiler_args = sorted(
            "%s=%s" % item for item in host_compiler_args.items()
        )

        platform_cmake_options = [
            "-GNinja",
            "-DCMAKE_MAKE_PROGRAM=%s" % api.ninja.path,
            "-DCMAKE_INSTALL_PREFIX=",
            "-DCMAKE_C_COMPILER=%s" % cipd_dir.joinpath("bin", "clang"),
            "-DCMAKE_CXX_COMPILER=%s" % cipd_dir.joinpath("bin", "clang++"),
            "-DCMAKE_ASM_COMPILER=%s" % cipd_dir.joinpath("bin", "clang"),
        ]
        if use_goma:  # pragma: no cover
            gomacc_path = api.goma.goma_dir / "gomacc"
            platform_cmake_options.extend(
                [
                    f"-DCMAKE_C_COMPILER_LAUNCHER={gomacc_path}",
                    f"-DCMAKE_CXX_COMPILER_LAUNCHER={gomacc_path}",
                    f"-DCMAKE_ASM_COMPILER_LAUNCHER={gomacc_path}",
                ]
            )

        platform_cmake_options.extend(
            {
                "linux": [
                    "-DCMAKE_AR=%s" % cipd_dir.joinpath("bin", "llvm-ar"),
                    "-DCMAKE_LINKER=%s" % cipd_dir.joinpath("bin", "ld.lld"),
                    "-DCMAKE_NM=%s" % cipd_dir.joinpath("bin", "llvm-nm"),
                    "-DCMAKE_OBJCOPY=%s" % cipd_dir.joinpath("bin", "llvm-objcopy"),
                    "-DCMAKE_OBJDUMP=%s" % cipd_dir.joinpath("bin", "llvm-objdump"),
                    "-DCMAKE_RANLIB=%s" % cipd_dir.joinpath("bin", "llvm-ranlib"),
                    "-DCMAKE_READELF=%s" % cipd_dir.joinpath("bin", "llvm-readelf"),
                    "-DCMAKE_STRIP=%s" % cipd_dir.joinpath("bin", "llvm-strip"),
                ],
                "mac": [
                    "-DCMAKE_LIBTOOL=%s"
                    % cipd_dir.joinpath("bin", "llvm-libtool-darwin"),
                    "-DCMAKE_LIPO=%s" % cipd_dir.joinpath("bin", "llvm-lipo"),
                ],
            }[api.platform.name]
        )
        platform_cmake_options.extend(
            [
                # TODO: once we upgrade to CMake 3.20, we can use
                # CMAKE_{C,CXX,ASM}_COMPILER_TARGET
                f"-DCMAKE_{lang}_FLAGS=--target={host_triple}"
                for lang in ["C", "CXX", "ASM"]
            ]
        )
        if api.platform.is_mac:
            platform_cmake_options.extend(["-DCMAKE_OSX_DEPLOYMENT_TARGET=10.13"])
            libcxx_path = cipd_dir.joinpath("lib", "libc++.a")
            platform_cmake_options.extend(
                [
                    f"-DCMAKE_{mode}_LINKER_FLAGS=-nostdlib++ {libcxx_path}"
                    for mode in ["SHARED", "MODULE", "EXE"]
                ]
            )

        with api.step.nest("zstd"):
            zstd_build_dir = api.path.start_dir / "zstd_build"
            api.file.ensure_directory("create zstd_build", zstd_build_dir)
            zstd_install_dir = api.path.start_dir / "zstd_install"
            api.file.ensure_directory("create zstd_install", zstd_install_dir)
            api.zstd.build(
                cmake_extra_args=platform_cmake_options + ["-DZSTD_BUILD_SHARED=OFF"],
                ninja_jobs=compile_jobs,
                build_dir=zstd_build_dir,
                install_dir=zstd_install_dir,
            )
        # These unfortunately must be in the environment for both configure and
        # make steps (because make runs more configure steps), rather than
        # being on the configure command line and getting saved that way.
        #
        # binutils wants just --with-zstd and uses pkg-cofnig if ZSTD_CFLAGS
        # and ZSTD_LIBS aren't set, which works with this environment.
        # GCC needs --with-zstd= with the path prefix to lib and include dirs.
        # So the configure arguments are plumbed individually below.
        zstd_env = {
            "PKG_CONFIG_PATH": f"{zstd_install_dir}/lib/pkgconfig",
            "PKG_CONFIG_SYSROOT_DIR": f"{zstd_install_dir}",
        }

        # We force LIMITS_H_TEST=true to tell the compiler to install a
        # <limits.h> that requires a system <limits.h> it can find via
        # #include_next.  But we don't have any system headers for the
        # just-built compiler building target code (i.e. libgcc).  So mock up a
        # little include directory that contains just an empty <limits.h> for
        # it to find.
        mock_include_dir = staging_dir / "mock-include"
        api.file.ensure_directory("create mock include dir", mock_include_dir)
        api.file.write_text(
            "write tmp <limits.h>", mock_include_dir / "limits.h", "/* tmp */"
        )

        with goma_context, api.context(
            env_prefixes={"PATH": [cipd_dir / "bin"]}, env=zstd_env
        ):
            for target, enable_targets in [
                ("aarch64", "arm-eabi"),
                ("riscv64", "riscv32-elf"),
                ("x86_64", "x86_64-pep"),
            ]:
                # configure arguments that are the same for binutils and gcc.
                common_args = host_compiler_args + [
                    "--disable-silent-rules",
                    "--prefix=",  # we're building a relocatable package
                    f"--target={target}-elf",
                    # Fuchsia uses .init/.fini arrays
                    "--enable-initfini-array",
                    # Good defaults for the tools.
                    "--enable-deterministic-archives",
                    "--enable-textrel-check=error",
                    # Zircon uses gold for userspace build
                    "--enable-gold",
                    # Enable plugins and threading for Gold.  This also happens
                    # to make it explicitly link in -lpthread and -dl, which
                    # are required by host_clang's static libc++.
                    "--enable-plugins",
                    "--enable-threads",
                    "--disable-werror",  # ignore warnings reported by Clang
                    "--disable-nls",  # no need for localization
                    "--with-included-gettext",  # use include gettext library
                ]

                # build binutils
                with api.step.nest(f"{target} binutils"):
                    binutils_build_dir = staging_dir.joinpath(
                        f"binutils_{target}_build_dir"
                    )
                    api.file.ensure_directory("create build dir", binutils_build_dir)

                    with api.context(cwd=binutils_build_dir):

                        def binutils_make_step(name, prefix, jobs, make_args=None):
                            # As of 2.32, gold/testsuite/Makefile.am
                            # unconditionally edits in a -B.../ switch to make
                            # the compiler use the just-built gold as the
                            # linker for test suite binaries.  This is wrong
                            # when building a cross-linker.  Force it to a
                            # no-op on the make command line to work around the
                            # bug.  TODO(mcgrathr): Drop this when we roll to a
                            # binutils that has this fixed upstream.
                            make_args = make_args or []
                            make_args.append("MAKEOVERRIDES=editcc=-eb")
                            return api.step(
                                name,
                                ["make", f"-j{jobs}"]
                                + make_args
                                + [
                                    f"{prefix}-{component}"
                                    for component in ["binutils", "gas", "ld", "gold"]
                                ],
                            )

                        try:
                            api.step(
                                "configure",
                                [
                                    binutils_dir / "configure",
                                    f"--with-pkgversion={binutils_pkgversion}",
                                    "--enable-deterministic-archives",  # more deterministic builds
                                    "--enable-textrel-check=error",
                                    f"--enable-targets={enable_targets}",
                                    "--with-zstd",
                                ]
                                + common_args,
                            )
                        except api.step.StepFailure as error:
                            log = api.file.read_text(
                                "config.log", binutils_build_dir / "config.log"
                            ).splitlines()
                            api.step.empty(
                                "binutils configure failure"
                            ).presentation.logs["config.log"] = log
                            raise error
                        try:
                            binutils_make_step("build", "all", compile_jobs)
                        except api.step.StepFailure as error:
                            logs = {
                                l[0]: api.file.read_text(
                                    "/".join(l), binutils_build_dir.joinpath(*l)
                                ).splitlines()
                                for l in [
                                    ("gas", "config.log"),
                                    ("binutils", "config.log"),
                                    ("ld", "config.log"),
                                    ("gold", "config.log"),
                                ]
                            }
                            step_result = api.step.empty("binutils build failure")
                            for name, text in sorted(logs.items()):
                                step_result.presentation.logs[name] = text
                            raise error
                        try:
                            binutils_make_step(
                                "test",
                                "check",
                                api.platform.cpu_count,
                                ["-k"]
                                # Gold's tests don't really work when the
                                # native target is not ELF, so skip them on
                                # macOS.
                                + (
                                    ["-o" "check-gold"]
                                    if api.platform.name == "mac"
                                    else []
                                ),
                            )
                        except api.step.StepFailure as error:
                            logs = {
                                l[0]: api.file.read_text(
                                    "/".join(l), binutils_build_dir.joinpath(*l)
                                ).splitlines()
                                for l in [
                                    ("gas", "testsuite", "gas.log"),
                                    ("binutils", "binutils.log"),
                                    ("ld", "ld.log"),
                                    ("gold", "testsuite", "test-suite.log"),
                                ]
                            }
                            step_result = api.step.empty("binutils test failure")
                            for name, text in sorted(logs.items()):
                                step_result.presentation.logs[name] = text
                            raise error
                        binutils_make_step(
                            "install", "install-strip", 1, [f"DESTDIR={pkg_dir}"]
                        )

                # build gcc
                with api.step.nest(f"{target} gcc"):
                    gcc_build_dir = staging_dir / f"gcc_{target}_build_dir"
                    api.file.ensure_directory("create build dir", gcc_build_dir)

                    with api.context(
                        cwd=gcc_build_dir, env_prefixes={"PATH": [pkg_dir / "bin"]}
                    ):
                        gcc_goals = ["gcc", "target-libgcc"]

                        def gcc_make_step(name, jobs, args, **kwargs):
                            cmd = [
                                "make",
                                f"-j{jobs}",
                                "MAKEOVERRIDES=USE_GCC_STDINT=provide LIMITS_H_TEST=true",
                                # Force flags for libgcc to get <limits.h> kludge.
                                f"CPPFLAGS_FOR_TARGET=-idirafter {mock_include_dir}",
                            ]
                            # Recipes won't let us capture stdout and stderr as
                            # a single stream, so use shell redirection to put
                            # stderr onto stdout.
                            cmd = [
                                "/bin/sh",
                                "-c",
                                " ".join(
                                    [pipes.quote(s) for s in cmd + args] + ["2>&1"]
                                ),
                            ]
                            return api.step(name, cmd, **kwargs)

                        api.step(
                            "configure",
                            [
                                gcc_dir / "configure",
                                f"--with-pkgversion={gcc_pkgversion}",
                                "--enable-languages=c,c++",
                                "--disable-lto",
                                # We don't need these runtime libraries.
                                "--disable-libstdcxx",
                                "--disable-libssp",
                                "--disable-libquadmath",
                                f"--with-zstd={zstd_install_dir}",
                            ]
                            + common_args
                            + (
                                # The runtimes must observe the Fuchsia ABI,
                                # which reserves x18.
                                ["CFLAGS_FOR_TARGET=-g -O2 -ffixed-x18"]
                                if target == "aarch64"
                                else []
                            ),
                        )
                        try:
                            gcc_make_step(
                                "build",
                                compile_jobs,
                                [f"all-{goal}" for goal in gcc_goals],
                                stdout=api.raw_io.output_text(
                                    name=f"{target} gcc build log", add_output_log=True
                                ),
                            )
                        finally:
                            logs = {
                                l[-1]: api.file.read_text(
                                    f"gcc {target} {'/'.join(l)}",
                                    gcc_build_dir.joinpath(*l),
                                ).splitlines()
                                for l in [
                                    ("gcc", "config.log"),
                                    ("gcc", "config.status"),
                                ]
                            }
                            step_result = api.step.empty("build logs")
                            for name, text in sorted(logs.items()):
                                step_result.presentation.logs[name] = text
                        # Skip tests on Mac because they are unreasonably slow.
                        if api.platform.name != "mac":
                            try:
                                gcc_make_step(
                                    "test", api.platform.cpu_count, ["check-gcc"]
                                )
                            finally:
                                logs = {
                                    l[-1]: api.file.read_text(
                                        f"gcc {target} {'/'.join(l)}",
                                        gcc_build_dir.joinpath(*l),
                                    ).splitlines()
                                    for l in [
                                        ("gcc", "testsuite", "gcc", "gcc.log"),
                                        ("gcc", "testsuite", "g++", "g++.log"),
                                    ]
                                }
                                step_result = api.step.empty("test logs")
                                for name, text in sorted(logs.items()):
                                    step_result.presentation.logs[name] = text
                        gcc_make_step(
                            "install",
                            1,
                            [f"DESTDIR={pkg_dir}"]
                            + [f"install-strip-{goal}" for goal in gcc_goals],
                        )

    binutils_version = api.file.read_text(
        "binutils version",
        binutils_dir.joinpath("bfd", "version.m4"),
        test_data="m4_define([BFD_VERSION], [2.27.0])",
    )
    m = re.match(r"m4_define\(\[BFD_VERSION\], \[([^]]+)\]\)", binutils_version)
    assert m and m.group(
        1
    ), f"bfd/version.m4 has unexpected format: {binutils_version!r}"
    binutils_version = m.group(1)
    gcc_version = api.file.read_text(
        "gcc version", gcc_dir.joinpath("gcc", "BASE-VER")
    ).rstrip()
    version = f"{gcc_version},{binutils_version}"

    # Copy the license file to the canonical name at the root of the package.
    api.file.copy(
        "copy license file", gcc_dir.joinpath("COPYING3"), pkg_dir.joinpath("LICENSE")
    )

    cipd_git_repository = f"{GCC_GIT},{BINUTILS_GIT}"
    cipd_git_revision = f"{gcc_revision},{binutils_revision}"

    # Add a bogus "salt" repository/revision to represent recipe changes when
    # the upstream revisions haven't changed.  Bump this number whenever
    # changing this recipe in ways that affect the package built without also
    # changing any upstream revision pin.
    # TODO(crbug.com/947158): Replace this with the recipes repo/rev when
    # infra makes that available here.
    if RECIPE_SALT:  # pragma: no cover
        cipd_git_repository += ",<salt>"
        cipd_git_revision += f",{RECIPE_SALT}"

    digest = api.cas_util.upload(pkg_dir, output_property="isolated")

    # The prod builders actually publish to CIPD.  The CI builders may run
    # quicker and they are the ones that trigger the test builds using the
    # toolchain just built.
    if prod:
        api.cipd_util.upload_package(
            "fuchsia/third_party/gcc/${platform}",
            pkg_dir,
            search_tag={"git_revision": cipd_git_revision},
            repository=cipd_git_repository,
            metadata=[("version", version)],
        )
    elif api.platform.name == "linux" and api.platform.arch == "intel":
        # Do a full integration build. This will use the just-built toolchain
        # to build all of Fuchsia to check whether there are any regressions.
        api.toolchain.trigger_build(
            "gcc_toolchain",
            GCC_GIT,
            gcc_revision,
            digest,
            git_repository_info=cipd_git_repository,
            git_revision_info=cipd_git_revision,
            builders=props.builders,
        )


def GenTests(api):
    binutils_revision = "3d861fdb826c2f5cf270dd5f585d0e6057e1bf4f"
    gcc_revision = "4b5e15daff8b54440e3fda451c318ad31e532fab"

    def test(
        name,
        platform="linux",
        arch="intel",
        bucket="ci",
        git_repo=GCC_GIT,
        revision=gcc_revision,
        fail=False,
    ):
        result = (
            api.buildbucket_util.test(
                name,
                bucket=bucket,
                git_repo=git_repo,
                revision=revision,
                status="FAILURE" if fail else "SUCCESS",
            )
            + api.platform.name(platform)
            + api.platform.arch(arch)
            + api.platform.bits(64)
            + api.properties(
                InputProperties(builders=["fuchsia/foo-arm64", "fuchsia/foo-x64"])
            )
        )
        if bucket == "ci" and not fail and platform == "linux" and arch == "intel":
            result += api.git.get_remote_branch_head("git ls-remote", "b" * 40)
        return result

    for platform, arch in (("linux", "intel"), ("linux", "arm"), ("mac", "intel")):
        yield (
            test(f"{platform}_{arch}", platform, arch)
            + api.step_data(
                "binutils-gdb.git rev-parse",
                api.raw_io.stream_output_text(binutils_revision),
            )
            + api.step_data(
                "gcc.git rev-parse", api.raw_io.stream_output_text(gcc_revision)
            )
        )
    yield (
        test("prod", bucket="prod")
        + api.step_data(
            "binutils-gdb.git rev-parse",
            api.raw_io.stream_output_text(binutils_revision),
        )
        + api.step_data(
            "gcc.git rev-parse", api.raw_io.stream_output_text(gcc_revision)
        )
    )
    yield (
        test(
            "binutils_configure_fail",
            git_repo=BINUTILS_GIT,
            revision=binutils_revision,
            fail=True,
        )
        + api.step_data(
            "gcc.git rev-parse", api.raw_io.stream_output_text(gcc_revision)
        )
        + api.step_data("x86_64 binutils.configure", retcode=1)
    )
    yield (
        test(
            "binutils_build_fail",
            git_repo=BINUTILS_GIT,
            revision=binutils_revision,
            fail=True,
        )
        + api.step_data(
            "gcc.git rev-parse", api.raw_io.stream_output_text(gcc_revision)
        )
        + api.step_data("x86_64 binutils.build", retcode=1)
    )
    yield (
        test(
            "binutils_test_fail",
            git_repo=BINUTILS_GIT,
            revision=binutils_revision,
            fail=True,
        )
        + api.step_data(
            "gcc.git rev-parse", api.raw_io.stream_output_text(gcc_revision)
        )
        + api.step_data("x86_64 binutils.test", retcode=1)
    )
    yield (
        test("gcc_test_fail", fail=True)
        + api.step_data(
            "binutils-gdb.git rev-parse",
            api.raw_io.stream_output_text(binutils_revision),
        )
        + api.step_data("aarch64 gcc.test", retcode=1)
    )
