# Copyright 2018 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Recipe for building LLVM."""

import contextlib
import re

from PB.recipes.fuchsia.contrib.llvm import InputProperties

DEPS = [
    "fuchsia/cas_util",
    "fuchsia/cipd_util",
    "fuchsia/cmake",
    "fuchsia/git_checkout",
    "fuchsia/gitiles",
    "fuchsia/goma",
    "fuchsia/linux_sdk",
    "fuchsia/macos_sdk",
    "fuchsia/ninja",
    "fuchsia/platform_util",
    "fuchsia/python3",
    "fuchsia/rbe",
    "fuchsia/zlib",
    "fuchsia/zstd",
    "recipe_engine/buildbucket",
    "recipe_engine/cipd",
    "recipe_engine/context",
    "recipe_engine/file",
    "recipe_engine/path",
    "recipe_engine/platform",
    "recipe_engine/properties",
    "recipe_engine/raw_io",
    "recipe_engine/step",
]

ARCH_TO_TARGET = {
    "x86_64": "x64",
    "aarch64": "arm64",
    "arm64": "arm64",
}

PLATFORMS = [
    "linux-amd64",
    "linux-arm64",
    "mac-amd64",
    "mac-arm64",
    "fuchsia-amd64",
    "fuchsia-arm64",
]

LLVM_PROJECT_GIT = "https://llvm.googlesource.com/llvm-project"

# The components to build & distribute. This contains all the dependencies of zxdb and Mesa,
# i.e. LLVMAArch64Disassembler, LLVMX86Disassembler, LLVMRISCVDisassembler, LLVMDebugInfoDWARF,
# and their recursive dependencies.
DEFAULT_COMPONENTS = [
    "llvm-headers",
    "LLVMAArch64Desc",
    "LLVMAArch64Disassembler",
    "LLVMAArch64Info",
    "LLVMAArch64Utils",
    "LLVMAnalysis",
    "LLVMAggressiveInstCombine",
    "LLVMAsmParser",
    "LLVMAsmPrinter",
    "LLVMBinaryFormat",
    "LLVMBitReader",
    "LLVMBitWriter",
    "LLVMBitstreamReader",
    "LLVMCFGuard",
    "LLVMCodeGen",
    "LLVMCodeGenTypes",
    "LLVMCore",
    "LLVMCoroutines",
    "LLVMDebugInfoBTF",
    "LLVMDebugInfoCodeView",
    "LLVMDebugInfoDWARF",
    "LLVMDebugInfoMSF",
    "LLVMDebugInfoPDB",
    "LLVMDemangle",
    "LLVMExecutionEngine",
    "LLVMFrontendOffloading",
    "LLVMFrontendOpenMP",
    "LLVMGlobalISel",
    "LLVMHipStdPar",
    "LLVMInstCombine",
    "LLVMInstrumentation",
    "LLVMipo",
    "LLVMIRPrinter",
    "LLVMIRReader",
    "LLVMLinker",
    "LLVMMC",
    "LLVMMCDisassembler",
    "LLVMMCJIT",
    "LLVMMCParser",
    "LLVMObjCARCOpts",
    "LLVMObject",
    "LLVMOrcShared",
    "LLVMOrcTargetProcess",
    "LLVMPasses",
    "LLVMProfileData",
    "LLVMRemarks",
    "LLVMRISCVDesc",
    "LLVMRISCVDisassembler",
    "LLVMRISCVInfo",
    "LLVMRuntimeDyld",
    "LLVMScalarOpts",
    "LLVMSelectionDAG",
    "LLVMSupport",
    "LLVMSymbolize",
    "LLVMTarget",
    "LLVMTargetParser",
    "LLVMTextAPI",
    "LLVMTransformUtils",
    "LLVMVectorize",
    "LLVMX86CodeGen",
    "LLVMX86Desc",
    "LLVMX86Disassembler",
    "LLVMX86Info",
]

PROPERTIES = InputProperties

FUCHSIA_REPO = "fuchsia.googlesource.com"
FUCHSIA_PROJECT = "fuchsia"
FUCHSIA_REF = "fc62c2617856ee41e37c04d08ef598afee6d69f7"
FUCHSIA_CMAKE_GEN_PATH = "scripts/clang/generate_cmake_sdk.py"
FUCHSIA_TEST_RUNNER_PATH = "scripts/clang/fuchsia_test_runner.py"

# TODO(https://fxbug.dev/327507421): Remove this temporary SDK version pin when
# we have a concrete solution.
FUCHSIA_SDK_VERSION = "f18"


def RunSteps(api, props):
    props.repository = props.repository or LLVM_PROJECT_GIT
    props.version = props.version or "integration"
    if not props.components:
        props.components.extend(DEFAULT_COMPONENTS)

    use_goma = (
        not api.platform.arch == "arm" and api.platform.bits == 64
    ) and not api.platform.is_win

    if props.use_rbe:
        ninja_jobs = api.goma.jobs  # doesn't really use Goma
        remote_build_context = api.rbe(
            absolute_path_policy=api.rbe.AbsolutePathPolicy.ALLOW
        )
    elif use_goma:
        ninja_jobs = api.goma.jobs
        remote_build_context = api.goma()
    else:
        ninja_jobs = api.platform.cpu_count
        remote_build_context = contextlib.nullcontext()

    host_platform = api.platform_util.host
    target_platform = api.platform_util.platform(
        props.platform or host_platform.platform
    )

    with api.step.nest("ensure_packages"):
        with api.context(infra_steps=True):
            cipd_dir = api.path.start_dir / "cipd"
            # TODO: deduplicate this and the clang toolchain recipe
            pkgs = api.cipd.EnsureFile()
            pkgs.add_package("fuchsia/third_party/clang/${platform}", props.version)
            if target_platform.is_fuchsia:
                pkgs.add_package(
                    "fuchsia/sdk/core/${platform}", FUCHSIA_SDK_VERSION, "sdk"
                )
            api.cipd.ensure(cipd_dir, pkgs)

    staging_dir = api.path.start_dir / "staging"
    pkg_dir = staging_dir / "root"
    api.file.ensure_directory("create pkg root dir", pkg_dir)

    llvm_dir, revision = api.git_checkout(props.repository, fallback_ref=props.revision)
    if target_platform.is_fuchsia:
        # TODO(fxbug.dev/136400): Replace it with Fuchsia SDK module.
        (
            fuchsia_test_dir,
            sdk_cmake_generator,
            fuchsia_test_runner,
        ) = fetch_fuchsia_sdk_helper_scripts(api)
        fuchsia_sdk_dir = cipd_dir / "sdk"

    # build llvm
    with api.macos_sdk(), api.linux_sdk(), remote_build_context:
        target_triple = target_platform.triple
        build_dir = staging_dir / "llvm_build_dir"
        api.file.ensure_directory("create llvm build dir", build_dir)
        target_arch = ARCH_TO_TARGET[target_triple.split("-")[0]]

        # Shared build options between mac and linux builders.
        base_options = [
            f"-DCMAKE_MAKE_PROGRAM={api.ninja.path}",
            "-DCMAKE_INSTALL_PREFIX=",
            f"-DCMAKE_C_COMPILER={cipd_dir.joinpath('bin', 'clang')}",
            f"-DCMAKE_CXX_COMPILER={cipd_dir.joinpath('bin', 'clang++')}",
            f"-DCMAKE_ASM_COMPILER={cipd_dir.joinpath('bin', 'clang')}",
        ]

        def add_platform_options(platform):
            options = []
            if not platform.is_mac:
                # Linux and Fuchsia shared options.
                options.extend(
                    [
                        f"-DCMAKE_AR={cipd_dir.joinpath('bin', 'llvm-ar')}",
                        f"-DCMAKE_LINKER={cipd_dir.joinpath('bin', 'ld.lld')}",
                        f"-DCMAKE_NM={cipd_dir.joinpath('bin', 'llvm-nm')}",
                        f"-DCMAKE_OBJCOPY={cipd_dir.joinpath('bin', 'llvm-objcopy')}",
                        f"-DCMAKE_OBJDUMP={cipd_dir.joinpath('bin', 'llvm-objdump')}",
                        f"-DCMAKE_RANLIB={cipd_dir.joinpath('bin', 'llvm-ranlib')}",
                        f"-DCMAKE_STRIP={cipd_dir.joinpath('bin', 'llvm-strip')}",
                    ]
                )

                if platform.is_fuchsia:
                    platform_sysroot = cipd_dir.joinpath(
                        "sdk", "arch", target_arch, "sysroot"
                    )
                    python3_path = api.python3._python3_path
                    options.extend(
                        [
                            f"-DLLVM_GTEST_RUN_UNDER={python3_path} {fuchsia_test_runner} run",
                            "-DLLVM_LIT_ARGS=-sv --no-gtest-sharding",
                        ]
                    )
                elif platform.is_linux:
                    platform_sysroot = api.linux_sdk.sysroot
            else:  # is_mac
                options.extend(
                    [
                        f"-DCMAKE_DSYMUTIL={cipd_dir.joinpath('bin', 'dsymutil')}",
                        f"-DCMAKE_LIBTOOL={cipd_dir.joinpath('bin', 'llvm-libtool-darwin')}",
                        f"-DCMAKE_LIPO={cipd_dir.joinpath('bin', 'llvm-lipo')}",
                    ]
                )
                # TODO(fxbug.dev/3043): Eventually use our own hermetic sysroot.
                platform_sysroot = api.macos_sdk.sysroot
            options.extend(
                [
                    f"-DCMAKE_SYSROOT={platform_sysroot}",
                    f"-DCMAKE_C_COMPILER_TARGET={platform.triple}",
                    f"-DCMAKE_CXX_COMPILER_TARGET={platform.triple}",
                    f"-DCMAKE_ASM_COMPILER_TARGET={platform.triple}",
                ]
            )
            compiler_launcher = None
            if props.use_rbe:
                compiler_launcher = api.rbe.cxx_compiler_wrapper()
            elif use_goma:
                compiler_launcher = api.goma.goma_dir / "gomacc"
            if compiler_launcher:
                options.extend(
                    [
                        f"-DCMAKE_C_COMPILER_LAUNCHER={compiler_launcher}",
                        f"-DCMAKE_CXX_COMPILER_LAUNCHER={compiler_launcher}",
                        f"-DCMAKE_ASM_COMPILER_LAUNCHER={compiler_launcher}",
                    ]
                )
            return options

        host_extra_options = base_options + add_platform_options(host_platform)
        target_extra_options = base_options + add_platform_options(target_platform)

        # TODO: deduplicate these flags with clang_toolchain recipe.
        if target_platform.is_linux:
            llvm_linker_flags = [
                f"-DCMAKE_{mode}_LINKER_FLAGS=-static-libstdc++ -ldl -lpthread"
                for mode in ["SHARED", "MODULE", "EXE"]
            ]
        elif target_platform.is_mac:
            llvm_linker_flags = [
                f"-DCMAKE_{mode}_LINKER_FLAGS=-nostdlib++ {cipd_dir.joinpath('lib', 'libc++.a')}"
                for mode in ["SHARED", "MODULE", "EXE"]
            ]
        elif target_platform.is_fuchsia:
            with api.step.nest("libzx"):
                libzx_build_dir = build_sdk_pkgs(
                    api,
                    pkgs=["zx"],
                    fuchsia_sdk_dir=fuchsia_sdk_dir,
                    gen_script_path=sdk_cmake_generator,
                    cmake_options=target_extra_options,
                )
            llvm_linker_flags = [
                f"-DCMAKE_C_FLAGS=-I{fuchsia_sdk_dir}/pkg/fdio/include -I{fuchsia_sdk_dir}/pkg/zx/include",
                f"-DCMAKE_CXX_FLAGS=-I{fuchsia_sdk_dir}/pkg/fdio/include -I{fuchsia_sdk_dir}/pkg/zx/include",
                f"-DCMAKE_SHARED_LINKER_FLAGS=-L{fuchsia_sdk_dir}/arch/{target_arch}/lib -L{libzx_build_dir} -ldl -lpthread -lfdio -lzircon -lzx",
                f"-DCMAKE_MODULE_LINKER_FLAGS=-L{fuchsia_sdk_dir}/arch/{target_arch}/lib -L{libzx_build_dir} -ldl -lpthread -lfdio -lzircon -lzx",
                f"-DCMAKE_EXE_LINKER_FLAGS=-L{fuchsia_sdk_dir}/arch/{target_arch}/lib -L{libzx_build_dir} -ldl -lpthread -lfdio -lzircon -lzx",
            ]

        # We only need a host build.
        with api.step.nest("zlib"):
            zlib_install_dir_host = api.zlib.build(
                cmake_extra_args=host_extra_options
                + [
                    # TODO: once we no longer build libLTO, we can drop
                    # this.
                    "-DCMAKE_POSITION_INDEPENDENT_CODE=ON",
                ],
                platform=host_platform,
                install_dir=api.path.mkdtemp("install"),
            )
            use_zlib_options = [
                f"-DZLIB_INCLUDE_DIR={zlib_install_dir_host / 'include'}",
                f"-DZLIB_LIBRARY={zlib_install_dir_host / 'lib' / ('zlibstatic.lib' if api.platform.is_win else 'libz.a')}",
            ]

        # We only need a host build.
        with api.step.nest("zstd"):
            zstd_install_dir_host = api.zstd.build(
                cmake_extra_args=host_extra_options
                + [
                    "-DZSTD_BUILD_SHARED=OFF",
                    # TODO: once we no longer build libLTO, we can drop
                    # this.
                    "-DCMAKE_POSITION_INDEPENDENT_CODE=ON",
                ],
                ninja_jobs=ninja_jobs,
                install_dir=api.path.mkdtemp("install"),
            )
            use_zstd_options = [
                f"-Dzstd_INCLUDE_DIR={zstd_install_dir_host / 'include'}",
                f"-Dzstd_LIBRARY={zstd_install_dir_host / 'lib' / 'libzstd.a'}",
            ]

        llvm_extra_options = [
            "-DCMAKE_BUILD_TYPE=RelWithDebInfo",
            f"-DLLVM_HOST_TRIPLE={host_platform.triple}",
            "-DLLVM_TARGETS_TO_BUILD=AArch64;RISCV;X86",
            "-DLLVM_DISTRIBUTION_COMPONENTS=" + ";".join(props.components),
            "-DLLVM_BUILD_LLVM_DYLIB=ON",
            "-DLLVM_ENABLE_PROJECTS=llvm",
            "-DLLVM_EXTERNALIZE_DEBUGINFO=ON",
            "-DLLVM_ENABLE_LIBXML2=OFF",
            "-DLLVM_ENABLE_TERMINFO=OFF",
            "-DLLVM_ENABLE_BACKTRACES=OFF",
            "-DLLVM_ENABLE_LLD=ON",
        ] + target_extra_options

        if target_platform != host_platform:
            # Doing LLVM cross compilation.
            system = target_platform.os.replace("mac", "darwin").capitalize()
            # Following options are needed for build/NATIVE step, which is responsible
            # for building host tools like TableGen during the cross compilation. If
            # these options are missing, the build bots will fail as they don't have
            # default toolchains like gcc.
            native_options = [
                "-DCMAKE_POLICY_DEFAULT_CMP0056=NEW",
                f"-DCMAKE_C_COMPILER={cipd_dir.joinpath('bin', 'clang')}",
                f"-DCMAKE_CXX_COMPILER={cipd_dir.joinpath('bin', 'clang++')}",
                f"-DCMAKE_ASM_COMPILER={cipd_dir.joinpath('bin', 'clang')}",
            ]
            if host_platform.is_linux:
                native_options.append(f"-DCMAKE_SYSROOT={api.linux_sdk.sysroot}")
            llvm_extra_options.extend(
                [
                    f"-DCMAKE_SYSTEM_NAME={system}",
                    f"-DCROSS_TOOLCHAIN_FLAGS_NATIVE={';'.join(native_options)}",
                ]
            )
            # TODO(fxbug.dev/125678): Enable ZLIB and ZSTD for Fuchsia cross
            # compilation when available.
        else:
            llvm_extra_options.extend(use_zlib_options)
            llvm_extra_options.extend(use_zstd_options)
            llvm_extra_options.extend(
                [
                    "-DLLVM_ENABLE_ZLIB=FORCE_ON",
                    "-DLLVM_ENABLE_ZSTD=FORCE_ON",
                ]
            )

        llvm_extra_options.extend(llvm_linker_flags)
        with api.step.nest(target_triple):
            api.cmake.build_with_ninja(
                src_dir=llvm_dir / "llvm",
                build_dir=build_dir,
                extra_args=llvm_extra_options,
                build_args=["distribution"],
                install_cmd="install-distribution",
                ninja_jobs=ninja_jobs,
                install_dir=pkg_dir,
            )

        if target_platform.is_fuchsia:
            with fuchsia_test_context(
                api,
                fuchsia_test_dir=fuchsia_test_dir,
                runner_script=fuchsia_test_runner,
                llvm_build_dir=build_dir,
                fuchsia_sdk_dir=fuchsia_sdk_dir,
                toolchain_dir=cipd_dir,
                fuchsia_target_triple=target_triple,
            ):
                try:
                    api.ninja(
                        "run unit tests",
                        build_dir=build_dir,
                        ninja_args=["check-llvm-unit"],
                    )
                except api.step.StepFailure as e:
                    if not props.ignore_test_failure:
                        raise e

    include_dir = pkg_dir / "include"
    step_result = api.file.read_text(
        "llvm-config.h",
        include_dir.joinpath("llvm", "Config", "llvm-config.h"),
        test_data='#define LLVM_VERSION_STRING "7.0.0svn"',
    )
    m = re.search(r'LLVM_VERSION_STRING "([a-zA-Z0-9.-]+)"', step_result)
    assert m, "Cannot determine LLVM version"
    llvm_version = m.group(1)

    api.file.copy(
        "copy license file", llvm_dir.joinpath("llvm", "LICENSE.TXT"), pkg_dir
    )

    api.cas_util.upload(pkg_dir)

    if props.upload_package:
        api.cipd_util.upload_package(
            f"fuchsia/third_party/llvm/{target_platform.platform}",
            pkg_dir,
            search_tag={"git_revision": revision},
            repository=props.repository,
            metadata=[("version", llvm_version)],
        )


@contextlib.contextmanager
def fuchsia_test_context(
    api,
    fuchsia_test_dir,
    runner_script,
    llvm_build_dir,
    fuchsia_sdk_dir,
    toolchain_dir,
    fuchsia_target_triple,
):
    with api.context(cwd=fuchsia_test_dir):
        api.python3(
            "start Fuchsia emulator",
            [
                runner_script,
                "start",
                "--build-dir",
                llvm_build_dir,
                "--sdk",
                fuchsia_sdk_dir,
                "--toolchain-dir",
                toolchain_dir,
                "--target",
                fuchsia_target_triple,
                "--local-product-bundle-path",
                fuchsia_test_dir / "local_pb",
                "--use-local-product-bundle-if-exists",
                "--verbose",
            ],
        )
    try:
        yield
    finally:
        try:
            api.file.read_text(
                "read emulator log", fuchsia_test_dir.joinpath("tmp~", "emulator_log")
            )
            api.file.read_text(
                "read emulator log",
                fuchsia_test_dir.joinpath("tmp~", "emulator_log.serial"),
            )
        except api.step.StepFailure:  # pragma: no cover
            # Ignore. Not a big deal.
            pass
        with api.context(cwd=fuchsia_test_dir):
            api.python3(
                "stop Fuchsia emulator",
                [
                    runner_script,
                    "stop",
                ],
            )


def fetch_fuchsia_sdk_helper_scripts(api):
    fuchsia_test_dir = api.path.start_dir / "fuchsia_test"
    sdk_cmake_generator = fuchsia_test_dir / "generate_cmake_sdk.py"
    fuchsia_test_runner = fuchsia_test_dir / "fuchsia_test_runner.py"
    api.file.ensure_directory("create fuchsia test directory", fuchsia_test_dir)
    script_data = api.gitiles.fetch(
        host=FUCHSIA_REPO,
        project=FUCHSIA_PROJECT,
        path=FUCHSIA_CMAKE_GEN_PATH,
        ref=FUCHSIA_REF,
        step_name="fetch sdk cmake generator",
        test_data=b"",
    )
    api.file.write_raw(
        "write sdk cmake generator file", sdk_cmake_generator, script_data
    )
    script_data = api.gitiles.fetch(
        host=FUCHSIA_REPO,
        project=FUCHSIA_PROJECT,
        path=FUCHSIA_TEST_RUNNER_PATH,
        ref=FUCHSIA_REF,
        step_name="fetch fuchsia test runner script",
        test_data=b"",
    )
    api.file.write_raw(
        "write fuchsia test runner script", fuchsia_test_runner, script_data
    )
    return (fuchsia_test_dir, sdk_cmake_generator, fuchsia_test_runner)


def build_sdk_pkgs(api, pkgs, fuchsia_sdk_dir, gen_script_path, cmake_options=()):
    sdk_build_dir = api.path.start_dir / "sdk_build"
    cmake_options = cmake_options[:]
    api.file.ensure_directory("create build dir for Fuchsia SDK", sdk_build_dir)
    cmake_output_path = fuchsia_sdk_dir / "CMakeLists.txt"
    gen_args = [
        gen_script_path,
        "--sdk-dir",
        fuchsia_sdk_dir,
        "--output",
        cmake_output_path,
    ]
    gen_args.extend(pkgs)
    api.python3(
        f"generate CMakeLists.txt for {pkgs}",
        gen_args,
    )
    api.cmake(
        src_dir=fuchsia_sdk_dir,
        build_dir=sdk_build_dir,
        extra_args=cmake_options,
    )
    api.ninja(
        "build",
        build_dir=sdk_build_dir,
    )
    return sdk_build_dir


def GenTests(api):
    revision = "75b05681239cb309a23fcb4f8864f177e5aa62da"
    for platform in PLATFORMS:
        host_os = "mac" if platform.startswith("mac") else "linux"
        if platform.startswith("fuchsia"):
            expect_fail = "arm64" in platform
            yield (
                api.test(
                    platform.replace("-", "_"),
                    status="FAILURE" if expect_fail else "SUCCESS",
                )
                + api.platform(host_os, 64)
                + api.buildbucket.ci_build(git_repo=LLVM_PROJECT_GIT, revision=revision)
                + api.properties(
                    platform=platform,
                    upload_package=True,
                    ignore_test_failure=not expect_fail,
                )
                + api.step_data(
                    "checkout.git rev-parse", api.raw_io.stream_output_text(revision)
                )
                + api.step_data("run unit tests", retcode=1)
            )
        else:
            yield (
                api.test(platform.replace("-", "_"))
                + api.platform(host_os, 64)
                + api.buildbucket.ci_build(git_repo=LLVM_PROJECT_GIT, revision=revision)
                + api.properties(platform=platform, upload_package=True)
                + api.step_data(
                    "checkout.git rev-parse", api.raw_io.stream_output_text(revision)
                )
            )

    yield (
        api.test("linux_arm64_rbe")
        + api.buildbucket.ci_build(git_repo=LLVM_PROJECT_GIT, revision=revision)
        + api.properties(platform="linux-arm64", use_rbe=True)
        + api.step_data(
            "checkout.git rev-parse", api.raw_io.stream_output_text(revision)
        )
    )
    yield (
        api.test("linux_arm64_host")
        + api.buildbucket.ci_build(git_repo=LLVM_PROJECT_GIT, revision=revision)
        + api.platform.name("linux")
        + api.platform.arch("arm")
        + api.platform.bits(64)
        + api.properties(platform="linux-arm64")
        + api.step_data(
            "checkout.git rev-parse", api.raw_io.stream_output_text(revision)
        )
    )
