# Copyright 2019 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Recipe for building and running LLVM test-suite."""

from PB.recipes.fuchsia.contrib.llvm_test_suite import InputProperties

DEPS = [
    "fuchsia/buildbucket_util",
    "fuchsia/cas_util",
    "fuchsia/cipd_util",
    "fuchsia/cmake",
    "fuchsia/git_checkout",
    "fuchsia/ninja",
    "recipe_engine/cipd",
    "recipe_engine/context",
    "recipe_engine/file",
    "recipe_engine/json",
    "recipe_engine/path",
    "recipe_engine/properties",
    "recipe_engine/step",
]

PROPERTIES = InputProperties


def RunSteps(api, props):
    target_platform = props.platform or api.cipd_util.platform_name

    with api.step.nest("cipd_packages"), api.context(infra_steps=True):
        cipd_dir = api.path.start_dir / "cipd"
        pkgs = api.cipd.EnsureFile()
        pkgs.add_package("fuchsia/third_party/clang/${platform}", "integration")
        pkgs.add_package(
            "fuchsia/third_party/sysroot/linux", "integration", target_platform
        )
        api.cipd.ensure(cipd_dir, pkgs)

    # Get the toolchain.
    with api.step.nest("toolchain"), api.context(infra_steps=True):
        clang_instance = props.clang_toolchain.instance
        clang_dir = api.path.mkdtemp("clang")

        if props.clang_toolchain.type == "cipd":
            pkgs = api.cipd.EnsureFile()
            pkgs.add_package("fuchsia/third_party/clang/${platform}", clang_instance)
            api.cipd.ensure(clang_dir, pkgs)
        elif props.clang_toolchain.type == "isolated":
            api.cas_util.download(digest=clang_instance, output_dir=clang_dir)
        else:  # pragma: no cover
            assert False, "unsupported clang_toolchain package type"

    test_suite_dir, _ = api.git_checkout(
        props.repository
        or "https://fuchsia.googlesource.com/third_party/llvm-test-suite"
    )

    build_dir = api.path.start_dir / "test-suite-build"
    api.file.ensure_directory("create test-suite-build dir", build_dir)

    with api.step.nest("test-suite"):
        api.cmake.build_with_ninja(
            src_dir=test_suite_dir,
            build_dir=build_dir,
            cache=test_suite_dir.joinpath("cmake", "caches", "O3.cmake"),
            extra_args=[
                f"-DCMAKE_MAKE_PROGRAM={api.ninja.path}",
                f"-DCMAKE_C_COMPILER={clang_dir.joinpath('bin', 'clang')}",
                f"-DCMAKE_CXX_COMPILER={clang_dir.joinpath('bin', 'clang++')}",
                "-DCMAKE_EXE_LINKER_FLAGS=-ldl -lpthread",
                "-DCMAKE_SHARED_LINKER_FLAGS=-ldl -lpthread",
                f"-DCMAKE_SYSROOT={cipd_dir / target_platform}",
                "-DTEST_SUITE_CXX_ABI=libc++",
            ],
        )
        api.step(
            "test",
            cmd=[
                "vpython3",
                api.resource("llvm-lit.py"),
                "-o",
                api.json.output(leak_to=build_dir / "results.json"),
                build_dir,
            ],
        )

    # TODO(fxbug.dev/3137): upload results.json to catapult


def GenTests(api):
    repo = "third_party/llvm-test-suite"
    yield (
        api.buildbucket_util.test("cipd", repo=repo)
        + api.properties(
            clang_toolchain={
                "type": "cipd",
                "instance": api.cipd.make_resolved_version(None),
            }
        )
    )
    yield (
        api.buildbucket_util.test("isolate", repo=repo)
        + api.properties(
            clang_toolchain={
                "type": "isolated",
                "instance": "abc123",
            },
        )
    )
