# Copyright 2023 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""A roller for timezone resources

Publishes a CIPD package based on the latest timezone data published
in the ICU data repository.
"""

# See here for the input parameters to this roller. They will become properties
# in the 'prop' parameter of RunSteps(api, props) below.
from PB.recipes.fuchsia.contrib.tzres_roller import InputProperties

DEPS = [
    "fuchsia/auto_roller",
    "fuchsia/cipd_resolver",
    "fuchsia/cipd_util",
    "fuchsia/git_checkout",
    "fuchsia/issuetracker",
    "recipe_engine/file",
    "recipe_engine/json",
    "recipe_engine/path",
    "recipe_engine/properties",
    "recipe_engine/step",
]

PROPERTIES = InputProperties

# The title of a bug to be filed.
BUG_SUMMARY = "action required: publish time zone update, version {version}"

# The body of a bug to be filed.
BUG_DESCRIPTION = """A new time zone data version was released: {version}.

These need to be rolled, built and patched into all current and future Fuchsia
releases.

----
This is a semi-automated process. It requires manual action, starting from the
point in the link below.

https://fuchsia.dev/internal/intree/development/update-time-zone-definitions?hl=en#2_update_pinned_versions

As the process becomes more automated, the manual part will shrink and possibly
disappear altogether.
"""


def RunSteps(api, props):
    # Check out timezone data.
    tmp_root_dir = api.path.cleanup_dir
    repo_dir, rev = api.git_checkout(repo=props.repo)
    tzdata_dirs = repo_dir.joinpath("tzdata", "icunew")
    # Directory names in tzdata_dirs are the tzdata versions we're after.
    # Let's find the latest one.
    tzdata_versions = []
    listdir = api.file.listdir(name="list tzdata version directory", source=tzdata_dirs)
    for one_dir in listdir:
        # We should check whether `dir` is a directory.
        tzdata_versions.append(api.path.basename(one_dir))

    # Determine the "latest" version.
    tzdata_versions.sort()
    latest_version = tzdata_versions[-1]
    api.step.empty("latest from git: {} at rev: {}".format(latest_version, rev))

    # The tag to use to build the tzdata CIPD package.
    tag_name = "version:{}".format(latest_version)
    package_name = props.tzdata_cipd_package_name
    cipd_package = api.cipd_resolver.resolve_common_tags(
        ref="latest",
        tag_name="version",
        packages=[package_name],
    )
    api.step.empty("from cipd: {}".format(cipd_package))

    # Check if we already have this version in CIPD.
    # If yes, nothing to roll.
    if tag_name in cipd_package:
        api.step.empty(
            "package {} with version {} already is in CIPD".format(
                package_name, latest_version
            )
        )
        return api.auto_roller.nothing_to_roll()

    # Build the cipd package.
    # Copy the files to a temporary directory.
    staging_dir = tmp_root_dir / "staging_dir"
    api.file.ensure_directory(
        name="create staging dir",
        dest=staging_dir,
    )

    license_file_src = repo_dir / "LICENSE"
    license_file_dest = staging_dir / "LICENSE"
    api.file.copy(
        name="staging LICENSE",
        source=license_file_src,
        dest=license_file_dest,
    )

    tzdata_dir = tzdata_dirs / latest_version
    for file in ["metaZones.res", "timezoneTypes.res", "zoneinfo64.res"]:
        src_file = tzdata_dir.joinpath(props.icu_data_version, props.endianness, file)
        dest_file = staging_dir / file

        api.file.copy(name="staging: {}".format(file), source=src_file, dest=dest_file)

    # upload the cipd package
    api.cipd_util.upload_package(
        pkg_name=props.tzdata_cipd_package_name,
        pkg_root=staging_dir,
        search_tag={"version": latest_version},
    )

    # file bug for updates
    api.issuetracker.file_bug(
        step_name="file IssueTracker bug",
        summary=BUG_SUMMARY.format(version=latest_version),
        description=BUG_DESCRIPTION.format(version=latest_version),
        # This component ID points to the top-level Fuchsia component.
        component_id=1360843,
    )
    return None


def GenTests(api):
    def properties(**kwargs):
        props = {
            "repo": "https://host/foo",
            "tzdata_cipd_package_name": "fuchsia/third_party/tzres",
            "icu_data_version": "44",
            "endianness": "le",
            "roll_options": {
                "remote": "sso://turquoise-internal/integration",
            },
        }
        props.update(**kwargs)
        return api.properties(**props)

    yield (
        api.test("nothing_to_roll")
        + properties()
        + api.step_data(
            "list tzdata version directory",
            api.file.listdir(["2022b", "2022", "2022a1", "2021"]),
        )
        + api.step_data("resolve common tags", api.json.output(["version:2022b"]))
    )

    yield (
        api.test("something_to_roll")
        + properties()
        + api.step_data(
            "list tzdata version directory",
            api.file.listdir(["2022b", "2022", "2022a1", "2021"]),
        )
        + api.step_data("resolve common tags", api.json.output(["version:2022a1"]))
    )
