# Copyright 2023 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Recipe for processing PGO data.

# Execution overview

## Get build results

This recipe gets or triggers profile builders and collects the results.

## Checkout + Build

It creates a checkout and builds the generated files so that these
can be uploaded with the profile data to GCS.

## Process Profile Data

It collects the profile data from all the builders and merges them into a single indexed file and uploads these to CAS and CIPD.

## Optimize Build

A separate pgo recipe will optimize using the profiles to better optimize compiled code.
"""


from PB.go.chromium.org.luci.buildbucket.proto import common as common_pb2
from PB.recipe_modules.recipe_engine.led.properties import (
    InputProperties as LedInputProperties,
)
from PB.recipes.fuchsia.fuchsia.profile import InputProperties

DEPS = [
    "fuchsia/build",
    "fuchsia/buildbucket_util",
    "fuchsia/cas_util",
    "fuchsia/checkout",
    "fuchsia/cipd_util",
    "fuchsia/subbuild",
    "fuchsia/testing",
    "recipe_engine/buildbucket",
    "recipe_engine/file",
    "recipe_engine/path",
    "recipe_engine/properties",
    "recipe_engine/step",
    "recipe_engine/swarming",
]

PROPERTIES = InputProperties

# The property name for the list of test task ids run by a fuchsia builder.
TASK_IDS_PROPERTY = "test-swarming-task-ids"


def RunSteps(api, props):
    build_ids = list(props.child_build_ids)
    if not props.child_build_ids:
        with api.step.nest("get builders to launch") as pres:
            builder_names = props.profile_builders
            if not builder_names:
                pres.step_text = "no builders specified"
                return
        build_ids = get_builds(
            api,
            builder_names,
        )

    assert build_ids, "failed to get builds"
    checkout = api.checkout.fuchsia_with_options(
        manifest=props.manifest, remote=props.remote
    )

    # Need to run api.build.gen() to generate the tool_paths.json to access the
    # llvm tools.
    gn_results = api.build.gen(
        checkout=checkout,
        fint_params_path=props.fint_params_path,
    )

    with api.step.nest("collect builds"):
        subbuilds = api.subbuild.collect(build_ids)
        builds = [s.build_proto for s in subbuilds.values()]

    profile_file_list = []
    for build in builds:
        task_ids = get_output_property(build, TASK_IDS_PROPERTY)
        if not task_ids:
            continue
        # All tasks should have been completed, so there's no need to have a timeout.
        results = api.swarming.collect(
            "collect",
            list(task_ids),
            output_dir=api.path.mkdtemp("swarming"),
        )
        for result in results:
            if not result.success:
                continue
            with api.step.nest(f"process result for {result.name}"):
                results_dir = api.testing.extract_test_results(
                    step_name="extract",
                    task_result=result,
                )
                raw_profiles = api.file.glob_paths(
                    "find raw profiles",
                    results_dir,
                    "**/*.profraw",
                    test_data=(["foo.profraw"]),
                )
                profile_file_list.extend(raw_profiles)

    # Upload the profile to cas/cipd
    with api.step.nest("merge profile data and upload to CAS and CIPD"):
        profile_dir = api.path.mkdtemp("profdata")
        profile_path = profile_dir / "fuchsia.profdata"
        # TODO(fxbug.dev/124414): Rust and C/C++ use different versions of llvm in fuchsia builds.
        # It's possible the profile data & llvm-profdata tool are incompatible between versions.
        # For now only C/C++ code is instrumented in the fuchsia profile variant.
        # When Rust profiling is enabled the data should be merged separately with the
        # corresponding Rust llvm-profdata tool.
        cmd = [
            gn_results.tool("llvm-profdata"),
            "merge",
            "-o",
            profile_path,
            "--failure-mode=any",
        ]
        cmd.extend(profile_file_list)
        api.step("llvm-profdata merge", cmd)

        if props.cipd_package:
            # TODO(fxbug.dev/124418): determine the required search tags
            api.cipd_util.upload_package(
                props.cipd_package, profile_path, {"pgo": "pgo"}
            )
        else:
            api.cas_util.upload(profile_dir, [profile_path])


def get_output_property(build, property_name):
    property_value = None
    if property_name in build.output.properties:
        property_value = build.output.properties[property_name]
    return property_value


def get_builds(api, builder_names):
    builders_to_launch = list(builder_names)
    with api.step.nest("launch builders") as presentation:
        builds = api.subbuild.launch(
            builders_to_launch,
            presentation,
            hide_in_gerrit=False,
        )
    return [build.build_id for build in builds.values()]


def GenTests(api):
    collect_steps = api.step_data(
        "collect",
        api.swarming.collect(
            [
                api.swarming.task_result(
                    id="610",
                    name="Linux",
                    outputs=["out/path/to/output/file"],
                ),
                api.swarming.task_result(
                    id="710",
                    failure=True,
                    name="QEMU",
                    outputs=["out/path/to/output/file"],
                ),
                api.swarming.task_result(
                    id="810",
                    name="QEMU-(2)",
                    outputs=["out/summary.json", "serial_log.txt"],
                ),
                api.swarming.task_result(
                    id="910",
                    name="QEMU-(3)",
                    outputs=["out/summary.json"],
                ),
            ]
        ),
    )

    def properties(**kwargs):
        props = {
            "manifest": "minimal",
            "remote": "https://fuchsia.googlesource.com/manifest",
            "fint_params_path": "fint_params/profile.textproto",
            "profile_builders": [
                "profile-builder",
                "profile-builder2",
                "profile-rust-builder",
            ],
            "artifact_gcs_bucket": "fuchsia-infra-artifacts",
        }
        props.update(kwargs)
        return api.properties(**props)

    profile_build = api.subbuild.ci_build_message(
        build_id=8945511751514863184,
        builder="profile-builder",
        output_props={
            "test-swarming-task-ids": ["610", "710", "810", "910"],
        },
        status="SUCCESS",
    )
    profile_led_build = api.subbuild.ci_build_message(
        builder="profile-builder",
        output_props={
            "test-swarming-task-ids": ["610", "710", "810", "910"],
        },
        input_props={
            "$recipe_engine/led": {
                "led_run_id": "led/user_example.com/abc123",
            },
        },
        status="SUCCESS",
    )
    profile_build_without_tasks = api.subbuild.ci_build_message(
        build_id=8945511751514863185,
        builder="profile-builder2",
        status="SUCCESS",
    )
    profile_build_with_infra_failure = api.subbuild.ci_build_message(
        build_id=8945511751514863186,
        builder="profile-builder3",
        status="INFRA_FAILURE",
    )
    profile_build_with_infra_failure.summary_markdown = "raised infra failure"
    profile_build_with_failure = api.subbuild.ci_build_message(
        build_id=8945511751514863187,
        builder="profile-builder4",
        status="FAILURE",
    )
    profile_build_with_failure.summary_markdown = "failed to build fuchsia"

    yield (
        api.buildbucket_util.test("default", tryjob=False)
        + properties(cipd_package="fuchsia/instrumentation/profile/test")
        + collect_steps
        + api.subbuild.child_build_steps(
            builds=[profile_build, profile_build_without_tasks],
            launch_step="launch builders",
            collect_step="collect builds",
        )
    )

    yield (
        api.buildbucket_util.test("default_cas", tryjob=False)
        + properties()
        + collect_steps
        + api.subbuild.child_build_steps(
            builds=[profile_build, profile_build_without_tasks],
            launch_step="launch builders",
            collect_step="collect builds",
        )
    )

    yield (
        api.buildbucket_util.test("empty_profile_builder_list", tryjob=False)
        + properties(profile_builders=[])
    )

    yield (
        api.buildbucket_util.test("default_led", tryjob=True)
        + properties(
            **{
                "$recipe_engine/led": LedInputProperties(
                    led_run_id="led/user_example.com/abc123",
                )
            }
        )
        + collect_steps
        + api.subbuild.child_led_steps(
            builds=[profile_led_build, profile_build_without_tasks],
            collect_step="collect builds",
        )
    )

    yield (
        api.buildbucket_util.test("led_with_child_build_ids", tryjob=True)
        + properties(
            child_build_ids=["8945511751514863184", "8945511751514863185"],
            **{
                "$recipe_engine/led": LedInputProperties(
                    led_run_id="led/user_example.com/abc123",
                )
            },
        )
        + collect_steps
        + api.buildbucket.simulated_collect_output(
            step_name="collect builds.collect",
            builds=[profile_build, profile_build_without_tasks],
        )
    )
