# Copyright 2021 The Fuchsia Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

"""Ensure the platform version in fuchsia.git is up-to-date.

The platform version should always be one ahead of the latest milestone branch's
milestone number.
"""

import json

from PB.go.chromium.org.luci.buildbucket.proto import common as common_pb2
from PB.recipe_engine.result import RawResult
from PB.recipes.fuchsia.release.platform_version_roller import InputProperties

DEPS = [
    "fuchsia/auto_roller",
    "fuchsia/build",
    "fuchsia/checkout",
    "fuchsia/git",
    "fuchsia/git_checkout",
    "fuchsia/gitiles",
    "fuchsia/release",
    "fuchsia/sso",
    "recipe_engine/context",
    "recipe_engine/properties",
    "recipe_engine/step",
]

PROPERTIES = InputProperties
BUMP_COMMIT_MESSAGE = """[roll] Introduce in-development API level {level}"""
FREEZE_COMMIT_MESSAGE = """[roll] Freeze API level {level}"""


def RunSteps(api, props):
    # This should be a minimal checkout only so that we can read the state of
    # the integration repository's branches.
    with api.step.nest("checkout integration"):
        integration_dir, _ = api.git_checkout(api.sso.sso_to_https(props.remote))

    with api.step.nest("query API levels"):
        # Use the Gitiles API to resolve the platform version, as this is much
        # faster than doing a full checkout. A full checkout will be performed
        # later if we realize that an update is needed.
        file_content = api.gitiles.fetch(
            "fuchsia.googlesource.com",
            "fuchsia",
            "sdk/version_history.json",
        ).decode()
        api_levels = json.loads(file_content)["data"]["api_levels"]
        all_api_levels = {int(api_level) for api_level in api_levels}

    if not all_api_levels:
        return RawResult(
            summary_markdown=f"No API levels found: {all_api_levels}",
            status=common_pb2.INFRA_FAILURE,
        )

    if not props.is_freeze:
        with api.step.nest("query latest milestone branch"), api.context(
            cwd=integration_dir
        ):
            milestones = query_milestones(api, props, integration_dir)

            if not milestones:
                return RawResult(
                    summary_markdown=f"{props.remote} does not have any milestone release branches",
                    status=common_pb2.INFRA_FAILURE,
                )

            latest_milestone_branch = max(milestones)

            # If running the recipe in dry_run mode, pretend there's a milestone
            # branch for current maximum API level. The CL won't be submitted.
            if props.roll_options.dry_run:
                with api.step.nest("override latest_milestone_branch for dry-run"):
                    latest_milestone_branch = max(all_api_levels)

        # If the latest milestone doesn't already have a corresponding API
        # level, something has gone wrong. Raise an exception and do not roll.
        if latest_milestone_branch not in all_api_levels:
            return RawResult(
                summary_markdown=f"""\
invalid platform versions: the latest milestone is {latest_milestone_branch},u
but that doesn't appear in the known API levels: {all_api_levels}""",
                status=common_pb2.INFRA_FAILURE,
            )

        desired_api_level = latest_milestone_branch + 1

        if desired_api_level in all_api_levels:
            api.step.empty("nothing to roll", step_text="platform version up-to-date")
            return api.auto_roller.nothing_to_roll()

    # At this point we've determined that we need to attempt a roll, so perform
    # a full checkout.
    checkout = api.checkout.fuchsia_with_options(
        manifest=props.manifest,
        remote=props.remote,
    )

    # Build Fuchsia to generate the golden files for API level N.
    # The build is expected to modify source files that are themselves build inputs,
    # so a subsequent build attempt is not expected to be clean.
    api.build.with_options(
        checkout=checkout,
        fint_params_path=props.fint_params_path,
        allow_dirty=True,
        clean_check=False,
    )

    if props.is_freeze:
        commit_message = FREEZE_COMMIT_MESSAGE.format(level=max(all_api_levels))
    else:
        commit_message = BUMP_COMMIT_MESSAGE.format(level=desired_api_level)

    change = api.auto_roller.attempt_roll(
        props.roll_options,
        repo_dir=checkout.root_dir,
        commit_message=commit_message,
    )
    return api.auto_roller.raw_result(change)


def query_milestones(api, props, integration_dir):
    api.git.fetch(all_remotes=True)
    branches = api.release.get_branches(repo_path=integration_dir, remotes=True)

    return [
        api.release.get_milestone_from_branch(b, props.branch_prefix) for b in branches
    ]


def GenTests(api):
    def properties(**kwargs):
        """Sample values of the roller's properties."""
        props = {
            "project": "integration",
            "manifest": "flower",
            "remote": "https://fuchsia.googlesource.com/integration",
            "branch_prefix": "f",
            "fint_params_path": "fint_params/platform-version-roller.textproto",
            "roll_options": {
                "remote": "sso://turquoise-internal/integration",
            },
        }
        props.update(**kwargs)
        return api.properties(**props)

    yield (
        api.test("no_api_levels", status="INFRA_FAILURE")
        + api.gitiles.fetch(
            "query API levels.fetch",
            b"""
    {
        "data": {
            "name": "Platform version map",
            "type": "version_history",
            "api_levels": {}
        },
        "schema_id": "https://fuchsia.dev/schema/version_history-22rnd667.json"
    }
            """,
        )
        + properties()
    )
    yield (
        api.test("no_branches", status="INFRA_FAILURE")
        + api.gitiles.fetch(
            "query API levels.fetch",
            b"""
    {
        "data": {
            "name": "Platform version map",
            "type": "version_history",
            "api_levels": {
                "2": {
                    "abi_revision": "0x099D5AB9C26B64DA",
                    "status": "in-development"
                }
            }
        },
        "schema_id": "https://fuchsia.dev/schema/version_history-22rnd667.json"
    }
            """,
        )
        + properties()
    )

    yield (
        api.test("roll")
        + api.release.get_branches(
            ["releases/foo", "releases/f1", "releases/f2"],
            nesting="query latest milestone branch",
        )
        + api.auto_roller.success()
        + api.gitiles.fetch(
            "query API levels.fetch",
            b"""
    {
        "data": {
            "name": "Platform version map",
            "type": "version_history",
            "api_levels": {
                "2": {
                    "abi_revision": "0x099D5AB9C26B64DA",
                    "status": "in-development"
                }
            }
        },
        "schema_id": "https://fuchsia.dev/schema/version_history-22rnd667.json"
    }
            """,
        )
        + properties()
    )

    yield (
        api.test("freeze")
        + api.auto_roller.success()
        + api.gitiles.fetch(
            "query API levels.fetch",
            b"""
    {
        "data": {
            "name": "Platform version map",
            "type": "version_history",
            "api_levels": {
                "2": {
                    "abi_revision": "0x099D5AB9C26B64DA",
                    "status": "in-development"
                }
            }
        },
        "schema_id": "https://fuchsia.dev/schema/version_history-22rnd667.json"
    }
            """,
        )
        + properties(is_freeze=True)
    )

    yield (
        api.test("nothing_to_roll")
        + api.release.get_branches(
            ["releases/foo", "releases/f1"],
            nesting="query latest milestone branch",
        )
        + api.gitiles.fetch(
            "query API levels.fetch",
            b"""
    {
        "data": {
            "name": "Platform version map",
            "type": "version_history",
            "api_levels": {
                "1": {
                    "abi_revision": "0x099D5AB9C26B64DA",
                    "status": "in-development"
                },
                "2": {
                    "abi_revision": "0x099D5AB9C26B64DA",
                    "status": "in-development"
                }
            }
        },
        "schema_id": "https://fuchsia.dev/schema/version_history-22rnd667.json"
    }
            """,
        )
        + properties()
    )

    yield (
        api.test("invalid_platform_version", status="INFRA_FAILURE")
        + api.release.get_branches(
            ["releases/foo", "releases/f1"],
            nesting="query latest milestone branch",
        )
        + api.gitiles.fetch(
            "query API levels.fetch",
            b"""
    {
        "data": {
            "name": "Platform version map",
            "type": "version_history",
            "api_levels": {
                "5": {
                    "abi_revision": "0x099D5AB9C26B64DA",
                    "status": "in-development"
                }
            }
        },
        "schema_id": "https://fuchsia.dev/schema/version_history-22rnd667.json"
    }
            """,
        )
        + properties()
    )

    yield (
        api.test("dry_run")
        + properties(roll_options={"dry_run": True})
        + api.release.get_branches(
            ["releases/foo", "releases/f1"],
            nesting="query latest milestone branch",
        )
        + api.gitiles.fetch(
            "query API levels.fetch",
            b"""
    {
        "data": {
            "name": "Platform version map",
            "type": "version_history",
            "api_levels": {
                "1": {
                    "abi_revision": "0x099D5AB9C26B64DA",
                    "status": "in-development"
                },
                "2": {
                    "abi_revision": "0x099D5AB9C26B64DA",
                    "status": "in-development"
                }
            }
        },
        "schema_id": "https://fuchsia.dev/schema/version_history-22rnd667.json"
    }
            """,
        )
        + api.auto_roller.dry_run_success()
    )
